set define '^'
set verify off
prompt ...htmldb_lang

create or replace package htmldb_lang is
--------------------------------------------------------------------------------
--
--  Copyright (c) 2003, 2023, Oracle and/or its affiliates.
--
--    NAME
--      htmldb_lang.sql
--
--    DESCRIPTION
--      This program allows for translation of text strings from
--      on national language to another.
--
--    MODIFIED   (MM/DD/YYYY)
--    cbackstr    10/14/2003 - Created
--
--------------------------------------------------------------------------------

function message (
    -- Function to return a message from the message repository.
    --
    -- p_name           - name of message to be printed
    -- p0 - p9          - substitution parameters that replace text srings
    --                    %0 through %9
    -- p_lang           - optional parameter to override the language
    -- p_application_id - optional parameter for a specific application ID instead of the current one
    --
    p_name                      in varchar2 default null,
    p0                          in varchar2 default null,
    p1                          in varchar2 default null,
    p2                          in varchar2 default null,
    p3                          in varchar2 default null,
    p4                          in varchar2 default null,
    p5                          in varchar2 default null,
    p6                          in varchar2 default null,
    p7                          in varchar2 default null,
    p8                          in varchar2 default null,
    p9                          in varchar2 default null,
    p_lang                      in varchar2 default null,
    p_application_id            in number default null)
    return varchar2
    ;



procedure message_p (
    -- Print a message from the message repository.
    --
    -- p_name           - name of message to be printed
    -- p0 - p9          - substitution parameters that replace text srings
    --                    %0 through %9
    -- p_lang           - optional parameter to override the language
    -- p_application_id - optional parameter for a specific application ID instead of the current one
    --
    p_name                      in varchar2 default null,
    p0                          in varchar2 default null,
    p1                          in varchar2 default null,
    p2                          in varchar2 default null,
    p3                          in varchar2 default null,
    p4                          in varchar2 default null,
    p5                          in varchar2 default null,
    p6                          in varchar2 default null,
    p7                          in varchar2 default null,
    p8                          in varchar2 default null,
    p9                          in varchar2 default null,
    p_lang                      in varchar2 default null,
    p_application_id            in number   default null)
    ;


function lang (
   -- Return a translated text string from the
   -- translatable messages repository within HTMLDB.
   --
   -- p_primary_text_string - text string to be translated
   -- p0 - p9  - substitution parameters that replace text srings
   --            %0 through %9
   -- p_primary_text_context-
   -- p_primary_language    -
   --
   p_primary_text_string       in varchar2 default null,
   p0                          in varchar2 default null,
   p1                          in varchar2 default null,
   p2                          in varchar2 default null,
   p3                          in varchar2 default null,
   p4                          in varchar2 default null,
   p5                          in varchar2 default null,
   p6                          in varchar2 default null,
   p7                          in varchar2 default null,
   p8                          in varchar2 default null,
   p9                          in varchar2 default null,
   p_primary_language          in varchar2 default null)
   return varchar2
   ;

procedure create_message(
    --
    -- Create a runtime message
    --
    -- p_application_id     - The ID of the application for which you wish to create the translatable text message. This is the ID of the primary language application.
    -- p_name               - The name of the translatable text message.
    -- p_language           - The IANA language code for the mapping. Examples include en-us, fr-ca, ja, he.
    -- p_message_text       - The text of the translatable text message.
    -- p_used_in_javascript - Specify if the message needs to be used directly by JavaScript code (using the apex.lang JavaScript API).
    --
    p_application_id        in number,
    p_name                  in varchar2,
    p_language              in varchar2,
    p_message_text          in varchar2,
    p_used_in_javascript    in boolean default false )
    ;

procedure update_message(
    --
    -- Update a text message specified by ID
    --
    p_id           in number,
    p_message_text in varchar2 );

procedure delete_message(
    --
    -- Delete a runtime message
    --
    p_id in number );


procedure update_translated_string(
    --
    -- Update the string in the translation repository
    --
    p_id       in number,
    p_language in varchar2,
    p_string   in varchar2);


procedure seed_translations(
    --
    -- Seed the translation repository for the specified application
    -- and language.
    --
    p_application_id in number,
    p_language       in varchar2 );

--==============================================================================
-- Create a language mapping for the specified application. A mapping
-- includes the target application ID and the language code (e.g., de, en-us, etc.)
--
-- Parameters:
-- * p_application_id:              The ID of the application for which you want to create the language mapping.
--                                  This is the ID of the primary language application.
-- * p_language:                    The IANA language code for the mapping. Examples include en-us, fr-ca, ja, he.
-- * p_translation_application_id:  Unique integer value for the ID of the underlying translated application.
--                                  This id must not end in zero.
-- * p_direction_right_to_left:     Specify document direction left-to-right or right-to-left.
-- * p_translation_image_directory: Specify the directory where images will be obtained.
--
-- Example:
-- The following example demonstrates the creation of the language mapping for an existing APEX application.
--
--    begin
--        --
--        -- If running from SQLcl, we need to set the environment
--        -- for the Oracle APEX workspace associated with this schema.
--        -- The call to apex_util.set_workspace is not necessary
--        -- if you're running within the context of the App Builder
--        -- or an APEX application.
--        --
--        apex_util.set_workspace( 'MY_WORKSPACE' );
--
--        -- Now, actually create the language mapping
--        apex_lang.create_language_mapping(
--            p_application_id => 100,
--            p_language => 'ja',
--            p_translation_application_id => 101,
--            p_direction_right_to_left => false,
--            p_image_directory => null );
--        commit;
--    end;
--==============================================================================
procedure create_language_mapping(
    p_application_id             in number,
    p_language                   in varchar2,
    p_translation_application_id in number,
    p_direction_right_to_left    in boolean  default false,
    p_image_directory            in varchar2 default null );

procedure update_language_mapping(
    --
    -- Update the mapping for specified application and language.  Only the mapping to the new translation
    -- ID can be updated.  The language cannot be updated.
    --
    p_application_id             in number,
    p_language                   in varchar2,
    p_new_trans_application_id   in number);


procedure delete_language_mapping(
    --
    -- Delete a language mapping for the specified application and language.
    --
    p_application_id in number,
    p_language       in varchar2);


--==============================================================================
-- Returns the XLIFF document for the specified language.
--
-- Parameters:
-- * p_application_id:         Application id of the primary application.
-- * p_page_id:                Optional page id if the XLIFF document should
--                             just contain the specified page.
-- * p_language:               The IANA language code for the existing translation
--                             mapping. Examples include en-us, fr-ca, ja, he.
-- * p_only_modified_elements: Choose to export all translatable elements
--                             of the application or only those elements
--                             which are new or have been updated.
--==============================================================================
function get_xliff_document (
    p_application_id         in number,
    p_page_id                in number  default null,
    p_language               in varchar2,
    p_only_modified_elements in boolean default false )
    return clob;

--==============================================================================
-- Applies the specified XLIFF document for the specified language to the
-- translation repository.
--
-- Parameters:
-- * p_application_id: Application id of the primary application.
-- * p_language:       The IANA language code for the existing translation
--                     mapping. Examples include en-us, fr-ca, ja, he.
-- * p_document:       The XLIFF document containing the translation.
--==============================================================================
procedure apply_xliff_document (
    p_application_id in number,
    p_language       in varchar2,
    p_document       in clob );


procedure publish_application(
    --
    -- Publish the translations of an application.  You can optionally specify
    -- a new application ID for the specific translation mapping.
    --
    p_application_id           in number,
    p_language                 in varchar2,
    p_new_trans_application_id in number default null );


function get_language_selector_list
   return varchar2;

procedure emit_language_selector_list;

end htmldb_lang;
/
show errors