set define '^' verify off
--------------------------------------------------------------------------------
--
--  Copyright (c) 2024, Oracle and/or its affiliates.
--
--  NAME
--    object_dependencies_views.sql
--
--  DESCRIPTION
--    Create views for Object Dependency reports
--
--  MODIFIED   (MM/DD/YYYY)
--  JPKEMP      02/19/2024 - Created
--
--------------------------------------------------------------------------------

prompt ... Create APEX Object Dependency views

-----------------------------------------------------
-- OBJECT DEPENDENCY
--

--------------------------------------------------------------------------------
--==============================================================================
prompt ...apex_used_db_object_comp_props
create or replace view apex_used_db_object_comp_props as
with q_default_group as (
    select wwv_flow_lang.system_message( title ) as group_name
    from   wwv_pe_display_groups
    where  name = 'SETTINGS' ),
q_comp_prop as (
    select cp.component_type_id,
           cp.property_id,
           dg.title as group_title,
           p.prompt
      from wwv_pe_component_props cp,
           wwv_pe_properties      p,
           wwv_pe_display_groups  dg -- OUTER
     where p.id     = cp.property_id
       and dg.id(+) = p.display_group_id ),
q_plug_prop as (
    select pa.id,
           pa.security_group_id,
           pa.prompt,
           coalesce(pg.title,
                    (select group_name from q_default_group) )  as group_title
      from wwv_flow_plugin_attributes  pa,
           wwv_flow_plugin_attr_groups pg -- OUTER
     where pg.id(+)                = pa.attribute_group_id
       and pg.security_group_id(+) = pa.security_group_id )
select f.workspace,
       f.workspace_display_name,
       ocp.flow_id                              as application_id,
       f.application_name,
       ocp.id,
       ocp.page_id,
       wwv_flow_lang.system_message( ct.singular_title )
                                                as component_type_name,
       ocp.component_id,
       ocp.component_display_name,
       case
         when q_plug_prop.id is not null
           then q_plug_prop.group_title
         when q_comp_prop.group_title is not null
           then wwv_flow_lang.system_message( q_comp_prop.group_title )
       end                                      as property_group_name,
       case
         when q_plug_prop.id is not null
           then q_plug_prop.prompt
         when q_comp_prop.prompt is not null
           then wwv_flow_lang.system_message( q_comp_prop.prompt )
       end                                      as property_name,
       ocp.code_fragment,
       ocp.error_message,
       ocp.analyzed_at,
       case when ocp.error_message is null then
         ( select count(*)
             from wwv_flow_used_obj_dependencies uod
            where ocp.id = uod.used_db_object_comp_prop_id )
       end                                      as object_dependency_count
  from wwv_flow_used_obj_comp_props  ocp
            join wwv_flow_authorized_extension f
                 on     f.workspace_id           = ocp.security_group_id
                    and f.application_id         = ocp.flow_id
            join wwv_pe_component_types        ct
                 on ct.id = ocp.component_type_id
                 -- each ocp should join to either a component or a plugin property:
       left join q_comp_prop
                 on     q_comp_prop.component_type_id = ocp.component_type_id
                    and q_comp_prop.property_id       = ocp.property_id
       left join q_plug_prop
                 on     q_plug_prop.id                = ocp.property_id
                    and q_plug_prop.security_group_id in ( ocp.security_group_id, 10)
/

begin
  wwv_flow_upgrade.add_table_comments (
    p_table_name        => 'apex_used_db_object_comp_props',
    p_table_comment     => 'Application code fragment analysed for database object dependencies',
    p_column_comments   => wwv_flow_t_varchar2 (
      'workspace'               , 'A work area mapped to one or more database schemas',
      'workspace_display_name'  , 'Display name for the workspace',
      'application_id'          , 'Application Primary Key, Unique over all workspaces',
      'application_name'        , 'Identifies the application',
      'id'                      , 'Identifies the code fragment',
      'component_display_name'  , 'Component name',
      'page_id'                 , 'Identifies page number',
      'component_type_name'     , 'Name of component type',
      'component_id'            , 'Primary key of the component, unique over all workspaces',
      'property_group_name'     , 'Property display group',
      'property_name'           , 'Name of component/plugin property',
      'code_fragment'           , 'The code fragment analysed for dependencies',
      'error_message'           , 'Compilation error (dependency analysis unable to be completed)',
      'analyzed_at'             , 'Date/time when this property was analyzed for dependencies',
      'object_dependency_count' , 'Count of objects this property references'
    )
  );
end;
/

--==============================================================================
prompt ...apex_used_db_objects
create or replace view apex_used_db_objects as
select f.workspace,
       f.workspace_display_name,
       uo.flow_id                  as application_id,
       f.application_name,
       uo.id,
       uo.referenced_type,
       uo.referenced_owner,
       uo.referenced_name,
       uo.referenced_sub_name,
       ( select count(*)
           from wwv_flow_used_obj_dependencies uod
          where uo.id = uod.used_db_object_id )
                                   as usage_count
  from wwv_flow_used_db_objects      uo,
       wwv_flow_authorized_extension f
 where f.workspace_id   = uo.security_group_id
   and f.application_id = uo.flow_id
/

begin
  wwv_flow_upgrade.add_table_comments (
    p_table_name        => 'apex_used_db_objects',
    p_table_comment     => 'Database object referenced by an application',
    p_column_comments   => wwv_flow_t_varchar2 (
      'workspace'              , 'A work area mapped to one or more database schemas',
      'workspace_display_name' , 'Display name for the workspace',
      'application_id'         , 'Application Primary Key, Unique over all workspaces',
      'application_name'       , 'Identifies the application',
      'id'                     , 'Identifies the database object',
      'referenced_type'        , 'Type of database object',
      'referenced_owner'       , 'Owner of the database object',
      'referenced_name'        , 'Name of database object',
      'referenced_sub_name'    , 'Depending on the referenced type may contain column name, function/procedure name or stay empty',
      'usage_count'            , 'Count of properties in the application that refer to this object'
    )
  );
end;
/

--==============================================================================
prompt ...apex_used_db_object_dependencies
create or replace view apex_used_db_obj_dependencies as
select f.workspace,
       f.workspace_display_name,
       uod.flow_id                  as application_id,
       f.application_name,
       ocp.page_id,
       uod.used_db_object_id,
       uo.referenced_type,
       uo.referenced_owner,
       uo.referenced_name,
       uo.referenced_sub_name,
       uod.used_db_object_comp_prop_id,
       ocp.component_type_name,
       ocp.component_display_name,
       ocp.property_group_name,
       ocp.property_name,
       ocp.code_fragment,
       ocp.analyzed_at
  from wwv_flow_used_obj_dependencies uod,
       wwv_flow_authorized_extension  f,
       wwv_flow_used_db_objects       uo,
       apex_used_db_object_comp_props ocp
 where f.workspace_id       = uod.security_group_id
   and f.application_id     = uod.flow_id
   and uo.id                = uod.used_db_object_id
   and ocp.id               = uod.used_db_object_comp_prop_id
   and uo.security_group_id = uod.security_group_id
/

begin
  wwv_flow_upgrade.add_table_comments (
    p_table_name        => 'apex_used_db_obj_dependencies',
    p_table_comment     => 'Database object referenced by an application component',
    p_column_comments   => wwv_flow_t_varchar2 (
      'workspace'                      , 'A work area mapped to one or more database schemas',
      'workspace_display_name'         , 'Display name for the workspace',
      'application_id'                 , 'Application Primary Key, Unique over all workspaces',
      'application_name'               , 'Identifies the application',
      'page_id'                        , 'Identifies page number',
      'used_db_object_id'              , 'Identifies the database object dependency',
      'referenced_type'                , 'Type of database object',
      'referenced_owner'               , 'Owner of the database object',
      'referenced_name'                , 'Name of database object',
      'referenced_sub_name'            , 'Depending on the referenced type may contain column name, function/procedure name or stay empty',
      'used_db_object_comp_prop_id'    , 'Identifies the application component property',
      'component_type_name'            , 'Name of component type',
      'component_display_name'         , 'Component name',
      'property_group_name'            , 'Property display group',
      'property_name'                  , 'Name of component/plugin property',
      'code_fragment'                  , 'The code fragment analysed for dependencies',
      'analyzed_at'                    , 'Date/time when this property was analyzed for dependencies'
    )
  );
end;
/

--------------------------------------------------------------------------------

prompt ...done creating APEX Object Dependency views

set define '^'
