set define off verify off
prompt ...wwv_flow_ai
create or replace package wwv_flow_ai as
--------------------------------------------------------------------------------
--
--  Copyright (c) 2024, Oracle and/or its affiliates
--
--    NAME
--      wwv_flow_ai.sql
--
--    DESCRIPTION
--      APEX Generative AI backend functionality.
--
--    SECURITY
--
--    NOTES:
--
--    RUNTIME DEPLOYMENT: YES
--
--    MODIFIED   (MM/DD/YYYY)
--      ralmuell  01/19/2024 - Created
--
--------------------------------------------------------------------------------

--==============================================================================
-- Globals
--==============================================================================
e_rag_no_tables_found        exception;
--==============================================================================
-- Types
--==============================================================================
subtype t_provider  is wwv_remote_servers.ai_provider_type%type;

type t_server is record (
    id                       number,
    name                     wwv_remote_servers.name%type,
    static_id                wwv_remote_servers.static_id%type,
    base_url                 wwv_remote_servers.base_url%type,
    --
    credential_id            wwv_remote_servers.credential_id%type,
    --
    ai_provider_type         wwv_remote_servers.ai_provider_type%type,
    ai_is_builder_service    wwv_remote_servers.ai_is_builder_service%type,
    ai_model_name            wwv_remote_servers.ai_model_name%type,
    ai_attributes            wwv_remote_servers.ai_attributes%type,
    ai_http_headers          wwv_remote_servers.ai_http_headers%type,
    --
    workspace_id             wwv_remote_servers.security_group_id%type );

--==============================================================================
-- Constants
--==============================================================================
c_provider_openai             constant    t_provider   := 'OPENAI';
c_provider_cohere             constant    t_provider   := 'COHERE';
c_provider_ocigenai           constant    t_provider   := 'OCI_GENAI';
c_provider_rest               constant    t_provider   := 'REST';

c_server_type_ai              constant    wwv_remote_servers.server_type%type := 'GENERATIVE_AI';

--==============================================================================
-- Gets the details of the server based either on the id or the static id.
-- Either p_id or p_static_id must be provided
--
-- Parameters:
-- * p_id               The AI service numeric id
-- * p_static_id        The AI service static id
--
-- Returns:
-- * The Remote Server
--==============================================================================
function get_server (
    p_id                in number   default null,
    p_static_id         in varchar2 default null )
    return t_server;

--==============================================================================
-- Are Generative AI features enabled for the current APEX Workspace
--
-- Returns:
-- * true if Generative AI features are enabled for the current workspace,
--   false otherwise
--==============================================================================
function is_enabled
return boolean;

--==============================================================================
-- Chat with a Generative AI service given a prompt and potential earlier
-- responses.
--
-- Parameters:
-- * p_prompt                The user prompt
-- * p_system_prompt         A (optional) system prompt to pass. Some Generative AI
--                           services (like OpenAI) support the use of passing a system
--                           prompt to set the context of a conversation
-- * p_service_id            The Generative AI Service ID
--                           If not provided, the app's default AI Service will be used.
-- * p_temperature           The temperature to use. How the temperature is interpreted depends
--                           on the Generative AI Serviice implementation. Higher temperatures
--                           result in more 'creative' responses.
-- * p_messages              The (optional) responses from an earlier conversation.
--                           Responses of procedure chat and nl2sql will be automatically
--                           added to p_responses for an easy conversational experience.
-- * p_application_id        The application ID, default is current application ID
--                           The application ID is used to set the proxy for the Application
--                           and to query application specific objects for prompt augmentation.
--
-- Returns:
-- * The response for the given prompt and type
--==============================================================================
function chat (
    p_prompt        in              varchar2,
    p_system_prompt in              varchar2                        default null,
    p_service_id    in              number                          default null,
    p_temperature   in              number                          default null,
    p_messages      in out nocopy   wwv_flow_ai_api.t_chat_messages )
    return clob;

--==============================================================================
-- Returns the AI consent message for the current application
--==============================================================================
function get_consent_message
return clob;

--==============================================================================
-- Generate a response for a given prompt.
--
-- Parameters:
-- * p_prompt                The user prompt
-- * p_service_id            The Generative AI Service static ID.
--                           If not provided, the app's default AI Service will be used.
-- * p_temperature           The temperature to use. How the temperature is interpreted depends
--                           on the Generative AI Service implementation. Higher temperatures
--                           result in more 'creative' responses. See the documentation of the
--                           Generative AI provider for details and allowed values.
--
-- Returns:
-- * The response for the given prompt and type
--
-- Example:
-- Generate a response with the configured Generative AI Service 'MY_AI_SERVICE' for
-- the given prompt
-- ~~~
-- declare
--   l_response clob;
-- begin
--   l_response := apex_ai.generate(
--     p_prompt            => 'What is Oracle APEX',
--     p_service_id        => 'MY_AI_SERVICE');
-- end;
-- ~~~
--==============================================================================
function generate (
    p_prompt            in              varchar2,
    p_service_id        in              number      default null,
    p_temperature       in              number      default null )
    return clob;

--==============================================================================
-- See wwv_flow_ai_api.is_user_consent_needed
--==============================================================================
function is_user_consent_needed (
    p_user_name         in  varchar2    default wwv_flow_security.g_user,
    p_application_id    in  number      default wwv_flow_security.g_flow_id )
    return boolean;

--==============================================================================
-- See wwv_flow_ai_api.set_user_consent
--==============================================================================
procedure set_user_consent (
    p_user_name         in  varchar2,
    p_application_id    in  number );

--==============================================================================
-- See wwv_flow_ai_api.revoke_user_consent
--==============================================================================
procedure revoke_user_consent (
    p_user_name         in  varchar2,
    p_application_id    in  number );

--==============================================================================
-- See wwv_flow_ai_api.revoke_user_consent_for_all
--==============================================================================
procedure revoke_user_consent_for_all (
    p_application_id    in  number );

--==============================================================================
procedure set_g_remote_server_id (
    p_remote_server_id  in  number );

--==============================================================================
-- handles component ajax requests
--==============================================================================
procedure ajax_component (
    p_enabled           boolean,
    p_remote_server_id  number,
    p_system_prompt     varchar2,
    p_plugin_name       varchar2 default null );

--==============================================================================
-- handles global ajax requests
--==============================================================================
procedure ajax_global;

--==============================================================================
-- emits the javascript dependencies of AI-enabled components
--==============================================================================
procedure emit_javascript_files;

end wwv_flow_ai;
/

show err
set define '^'
