/*!
 * Copyright (c) 2023, Oracle and/or its affiliates.
 */

/**
 * This file initializes the "gallery" of the diagram builder. Currently, it only contains the list of available activities of the worfklow version.
 * Main function of this file is to create the IconList and get the data of the selected component before drag and let the diagram know that a drag has started.
 * All other processses related to the drag-and-drop are controlled by the diagram.
 * 
 * 
 */

/* global pe */

window.componentGallery = {};

( function( model, $, util, pd ) {
    "use strict";

    const hasOwnProperty = util.hasOwnProperty;
    const ACTIVITY = "activity";
    const gComponents = { activities: [] }, 
        gTypeToClassMap = { "activity" : "a-Gallery-activity" },
        gTypeToCollectionMap = { [ ACTIVITY ]: "activities" };

    function renderLibComponent( out, index, item ) {
        let className = gTypeToClassMap[ item.type ],
            isPageReadOnly = model.isPageReadOnly();

        if ( !isPageReadOnly ) {
            className += " is-draggable";
        }

        out.markup( "<li" )
            .attr( "class", className )
            .optionalAttr( "aria-grabbed", isPageReadOnly ? null : "false" )
            .attr( "data-component", item.type + ":" + index )
            .attr( "draggable", "true" )
            .markup( "><span class='a-Icon " ).attr( item.iconClass ).markup( "'></span><span class='a-Gallery-componentName'>" ).content( item.title )
            .markup( "</span></li>" );
    }

    function getComponentData( key ) {
        let data,
            parts = key.split(":" ),
            collection = gComponents[ gTypeToCollectionMap[ parts[ 0 ] ] ];

        if ( collection ) {
            data = collection[ parts[ 1 ] ];
        }

        if ( !data ) {
            throw new Error( "Invalid component data key" );
        }
        return data;
    }

    function updateComponentGallery() {
        let diagram = $( "#app" )[ 0 ].diagramBuilder,
            isPageReadOnly = model.isPageReadOnly(),
            out = util.htmlBuilder();

        [ "activities" ].forEach( function( value ) {
            let item, components = gComponents[ value ],
                list$ = $( "#cg-" + util.escapeCSS( value ) + " ul" ).first();

            out.clear();
            for ( let i = 0; i < components.length; i++ ) {
                item = components[i];
                if ( item.isSupported ) {
                    renderLibComponent( out, i, item );
                }
            }
            list$.html( out.toString() )
                .iconList( "refresh" );

            // make component icons draggable but only if the page is not readonly
            if ( !isPageReadOnly ) {
                $( "#cg-" + util.escapeCSS( value ) + " li" ).draggable( {
                    appendTo: "#app-viewport", // append here for proper scrolling
                    // cursor2: "default",
                    containment: "#glv-viewport-container",
                    scroll: false,
                    cursorAt: { left: 2, top: 2 },
                    delay: 10,
                    distance: 5,
                    opacity: 0.1,
                    zIndex: 1000,
                    addClasses: false,
                    revert: false,
                    helper: function() {
                        return document.createElement( "div" );
                    },
                    start: function ( event ) {
                        let comp$ = $( this ),
                            key = comp$.attr( "data-component" ),
                            comp = getComponentData( key ),
                            version = diagram.getCurrentVersion(),
                            elementType = comp.values.find( val => val.id === model.PROP.PAGE_PROCESS_TYPE );

                        // stop if the diagram is read-only or the source can't be modified of any of the links ( happens in active state )
                        if ( diagram.isReadOnly() || ( version && version.getProperty( model.PROP.STATE )?.getValue() !== "DEVELOPMENT" ) ) {
                            return false;
                        }
                        diagram.startDragging( elementType.value, event );
                        apex.tooltipManager.disableTooltips();
                    },
                    stop: function ( /* event, ui  */) {
                        apex.tooltipManager.enableTooltips();
                    }
                } );
            }
        } );
    }

    function initComponentGallery() {
        let components = pd.getWorkflowActivitiesConfig();
        for ( let act in components ) {
            if ( hasOwnProperty( components, act ) ) {
                gComponents.activities.push( {
                    type: ACTIVITY,
                    iconClass: pd.getComponentIconClass( ACTIVITY, act ),
                    title: components[ act ].title,
                    isSupported: true,
                    values: [
                        { id: model.PROP.PAGE_PROCESS_TYPE, value: act }
                    ]
                } );
            }
        }

        updateComponentGallery();
    }

    $( function() { 

        $( "#cg-activities ul" ).addClass( "a-Gallery" ).each( function() {
            $( this ).iconList({
                multiple: false
            });
        } );

        $( "#gallery-tabs .u-ScrollingViewport.resize" ).resize( function() {
            $( this ).children( ".a-IconList" ).resize(); // fire resize event without actually changing the size of the list.
        });

    } );

    $( document ).on( "diagramReady", function() {
        initComponentGallery();
    } );


})( pe, apex.jQuery, apex.util, window.pageDesigner );