/*!
 * Copyright (c) 2023, 2024, Oracle and/or its affiliates.
 */

/* global tinymce */
/* eslint quotes: ["error", "double"] */
/* eslint no-var: "error" */
/* eslint "comma-dangle": ["error", "never"] */
/* eslint keyword-spacing: ["error", { "before": true, "after": true } ] */

/**
 *  A Web Component implementation of a Rich Text Editor based on TinyMCE.
 *
 *  Dependencies:
 *      - dompurify
 *      - tinymce
 *          - include all plugins for faster load times
 *      - markedjs (for markdown)
 *      - turndown (for markdown)
 *
 *  Minimap markup:
 *      <a-rich-text-editor id="P1_ITEM"></a-rich-text-editor>
 *
 *
 * Initial Content:
 *      The widget can be prerendered with the final content, for smoother loading, as such:
 *          <a-rich-text-editor id="P1_ITEM">
 *              <div class="a-rich-text-editor--loading-content mce-content-body">
 *                  <!-- whatever content -->
 *              </div>
 *          </a-rich-text-editor>
 *      This inner element will be hidden once the editor is fully loaded.
 *      This is currently only doable in markdown mode, as the server can prerender this content.
 *      In HTML mode, the server cannot sanitize this content, so this element is not rendered.
 *
 *  Editor Modes:
 *      The editor has 2 modes: HTML and Markdown.
 *      In HTML mode, the editor is fed HTML, and it returns HTML.
 *      In Markdown mode, the editor is fed Markdown, and it returns Markdown.
 *          TinyMCE however internally only works with HTML, so for Markdown mode, we need to
 *          upcast from, and downcast to HTML, using MarkedJS and TurndownJS respectively.
 *      Specify the mode attribute to choose. Markdown is the default.
 *      <a-rich-text-editor id="P1_ITEM" mode="html"></a-rich-text-editor>
 *      <a-rich-text-editor id="P1_ITEM" mode="markdown"></a-rich-text-editor>
 *
 *  Value:
 *      The initial value can be passed via the value attribute of the web component:
 *          <a-rich-text-editor id="P1_ITEM" value="## heading\n\n> blockquote\n"></a-rich-text-editor>
 *
 *
 *  Final Markup for an HTML editor:
 *      <a-rich-text-editor id="P1_ITEM" mode="html">
 *          <input type="hidden">
 *          <div class="a-rich-text-editor--dummy" data-editor="P1_ITEM" style="display: none"></div>
 *          <div role="application"> <!-- the actual widget -->
 *              ...
 *          </div>
 *      </a-rich-text-editor>
 *
 *  Final Markup for a Markdown editor, with an initial value as example:
 *      <a-rich-text-editor id="P1_ITEM" mode="markdown" value="## heading\n\n> quote\n">
 *          <input type="hidden">
 *          <div class="a-rich-text-editor--dummy" data-editor="P1_ITEM" style="display: none"></div>
 *          <div role="application"> <!-- the actual widget -->
 *              ...
 *          </div>
 *      </a-rich-text-editor>
 *
 *  Interactive Grid integration:
 *      When an editable cell is activated in the IG, the page items is moved around in the DOM,
 *      and is either placed directly in the cell, or in a draggable dialog.
 *      While this works for all other page items, TinyMCE uses an iframe under the hood which
 *      cannot be moved around in the DOM. Attempts were made to reload it on each cell activation,
 *      but the results were not adequate. Therefore, this Web Component offers 2 visual modes:
 *          1) inline, default
 *              <a-rich-text-editor id="P1_ITEM" visual-mode="inline"></a-rich-text-editor>
 *          2) dialog, meant for IG use
 *              <a-rich-text-editor id="P1_ITEM" visual-mode="dialog"></a-rich-text-editor>
 *
 *      In dialog mode, the widget only displays 1 readonly field, and a button, similar to the PopupLOV.
 *      The readonly field is kept up to date with the item's value.
 *      In this mode, the editor is placed at the bottom of the page, within a dialog.
 *      When the button is pressed, the dialog is open, and the user can edit the value.
 *      This way, the editor will always stay within the same place, and is not moved around.
 *
 *      Final markup of a dialog editor:
 *
 *          <a-rich-text-editor id="P1_ITEM" visual-mode="dialog">
 *              <input type="hidden">
 *              <!-- note. preview and button markup still to change -->
 *              <div class="a-rich-text-editor--preview">
 *                  <!-- value updated as the editor is updated -->
 *              </div>
 *              <button class="a-rich-text-editor--button">
 *              </button>
 *          </a-rich-text-editor>
 *
 *          <div class="a-rich-text-editor--dialog">   <!-- jquery dialog -->
 *              <div class="a-rich-text-editor--dummy" data-editor="P1_ITEM" style="display: none"></div>
 *              <div role="application"> <!-- the actual widget -->
 *                  ...
 *              </div>
 *          </div>
 *
 *  Internal. Getting and Setting the value:
 *      - The public entry points to get and set the editor values are getValue and setValue
 *      - Internally, the single source of truth is the value of _wrappedElement
 *      - The editor will try to keep the value of _wrappedElement in sync as appropriate
 *          - note it won't do this on each key stroke. it employs a comparison between
 *              a temporary value stored on focusin and the value on focusout
 *          - further, in dialog mode, the value is only fully set once the user hit OK
 *      - In dialog mode, the preview element is kept in sync with the value of _wrappedElement
 *      - To achieve all this, we make use of the helper function:
 *          - #getTransientValue, #setTransientValue and #commitTransientValue
 *
 *  Restrictions:
 *      - currently, the editor does not re-trigger events such as click, keyup, etc
 *          that can be listened to by dynamic actions. one should use the TinyMCE
 *          instance by invoking getEditor() to access these internal events
 */

( function( item, util, debug, $, WebComponent, lang, theme ) {
    "use strict";

    const trace = ( ...args ) => {
        debug.trace( "a-rich-text-editor", ...args );
    };

    const msg = key => {
        return lang.getMessage( "APEX.RTE." + key );
    };

    const keyCodes = $.ui.keyCode;

    const CLS_IS_FOCUSED = "is-focused";
    const CLS_PLACEHOLDER = "apex-item-placeholder";
    const CLS_DUMMY = "a-rich-text-editor--dummy";
    const CLS_PREVIEW = "a-rich-text-editor--preview";
    const CLS_PREVIEW_RICH_TEXT = "a-rich-text-editor--preview-rich-text";
    const CLS_BUTTON = "a-rich-text-editor--button";
    const CLS_DIALOG = "a-rich-text-editor--dialog";

    const POPUP_SELECTOR = ".tox, ." + CLS_DIALOG;

    const MODE_HTML = "html";
    const MODE_MARKDOWN = "markdown";

    // for html mode only
    const TOOLBAR_BASIC = "basic";
    const TOOLBAR_INTERMEDIATE = "intermediate";
    const TOOLBAR_FULL = "full";

    // for html mode only
    const TOOLBAR_STYLE_OVERFLOW = "overflow";
    const TOOLBAR_STYLE_MULTILINE = "multiline";

    // regular page item or ig column
    const VISUAL_MODE_INLINE = "inline";
    const VISUAL_MODE_DIALOG = "dialog";

    // for ig columns only
    const DISPLAY_VALUE_RICH_TEXT = "rich-text";
    const DISPLAY_VALUE_ESCAPED = "escaped";
    const DISPLAY_VALUE_PLAIN_TEXT = "plain-text";

    // text messages
    const MSG_TOOLBAR_FORMATTING = msg( "TOOLBAR_FORMATTING" );
    const MSG_TOOLBAR_ALIGNMENT = msg( "TOOLBAR_ALIGNMENT" );
    const MSG_TOOLBAR_LISTS = msg( "TOOLBAR_LISTS" );
    const MSG_TOOLBAR_FONT= msg( "TOOLBAR_FONT" );
    const MSG_TOOLBAR_EXTRAS= msg( "TOOLBAR_EXTRAS" );
    const MSG_RICH_TEXT_EDITOR = msg( "RICH_TEXT_EDITOR" );
    const MSG_READ_ONLY_RICH_TEXT_EDITOR = msg( "READ_ONLY_RICH_TEXT_EDITOR" );

    // tinymce specific constants
    const TINY_ENABLED = "design";
    const TINY_DISABLED = "readonly";

    const DEFAULT_MIN_HEIGHT = 180;

    const CONTENT_STYLE = ( () => {
        const vars = [
            "--a-base-font-family",
            "--a-base-font-weight-semibold",
            "--a-field-input-text-color",
            "--a-field-input-background-color",
            "--a-field-input-font-size",
            "--a-field-input-hover-text-color",
            "--a-field-input-hover-background-color",
            "--a-field-input-focus-text-color",
            "--a-field-input-focus-background-color",
            "--a-field-input-padding-y",
            "--a-field-input-padding-x",
            "--a-field-input-border-width",
            "--a-field-input-transition",
            "--a-gv-border-color",
            // prism variables
            "--prism-comment",
            "--prism-string",
            "--prism-literal",
            "--prism-keyword",
            "--prism-function",
            "--prism-deleted",
            "--prism-class",
            "--prism-builtin",
            "--prism-property",
            "--prism-namespace",
            "--prism-number",
            "--prism-decorator",
            "--prism-boolean",
            "--prism-variable",
            "--prism-constant",
            "--prism-symbol",
            "--prism-interpolation",
            "--prism-selector",
            "--prism-keyword",
            "--prism-regex",
            "--prism-json",
            "--prism-inline"
        ];

        const style = getComputedStyle( document.documentElement );
        const getRule = name => `${name}:${style.getPropertyValue( name ) || "unset" };`;

        return [
            ":root {",
            ...( vars.map( v => "    " + getRule( v ) ) ),
            "}"
        ].join( "\n" );
    } )();

    /**
     * Delay the global configuration of dependencies until the first editor that needs them loads.
     * Do this just in case the dependencies are loaded after this file.
     * This might happen if there are 2 RTEs on the page, the first HTML and the second MD.
     */
    let markdownDependenciesConfigured = false,
        turndownService;

    const configureMarkdownDependencies = () => {
        // turndown
        turndownService = new window.TurndownService( {
            codeBlockStyle: "fenced",
            hr: "---",
            headingStyle: "atx",
            // we wish to preserve empty paragraphs, which can be used to insert blank lines.
            // while we don't allow for any other type of custom HTML, this is an exception.
            // apex_markdown.to_html will likewise let <p></p> pass through, regardless of embedded_html_mode.
            // without this custom option, <p></p> would get removed.
            // see 35308544
            blankReplacement( content, node ) {
                if ( node.nodeName === "P" ) {
                    return "<p></p>";
                }
            }
        } );
        window.turndownService = turndownService;
        // marked
        // it appears these options cannot be passed to marked.parse directly, so we're declaring them globally
        // this code originated from the ckeditor5 approach to Markdown, so we're keeping it as before
        window.marked.use( {
            tokenizer: {
                // Disable the autolink rule in the lexer.
                autolink: () => null,
                url: () => null
            },
            renderer: {
                checkbox( ...args ) {
                    // Remove bogus space after <input type="checkbox"> because it would be preserved
                    // by DomConverter as it's next to an inline object.
                    return Object.getPrototypeOf( this ).checkbox.call( this, ...args ).trimRight();
                },
                code( ...args ) {
                    // Since marked v1.2.8, every <code> gets a trailing "\n" whether it originally
                    // ended with one or not (see https://github.com/markedjs/marked/issues/1884 to learn why).
                    return Object.getPrototypeOf( this ).code.call( this, ...args ).replace( "\n</code>", "</code>" );
                }
            }
        } );
    };

    const getDefaultEditorOptions = ( mode, toolbarVariant = TOOLBAR_BASIC, toolbarStyle = TOOLBAR_STYLE_OVERFLOW ) => {
        const options = {
            menubar: false,
            promotion: false,   // remove upgrade button on the top
            branding: false,    // remove tinymce logo in the footer
            // toolbar
            toolbar_location: "top",
            // if the mobile object is not specified, on mobile, the toolbar_mode will default to "scrolling"
            mobile: {
                toolbar_mode: "sliding"
            },
            // keep the plugins in sync with the serverside render function to speed up the initial load
            plugins: [
                "lists",
                "table",
                "link",
                "code",
                "image",
                "pagebreak",
                "codesample-pre",
                "codesample",
                "codesample-post",
                "preview",
                "autoresize"
            ].join( " " ),
            // plugin specific settings:
            lists_indent_on_tab: false, // would otherwise cause a tab trap
            //
            statusbar: false,
            hidden_input: false,
            // only show a context menu when right clicking on an image
            // no context menu for links which would always appear on right click
            contextmenu: "image",
            // styles start
            width: "100%",
            content_css: apex.env.APEX_FILES + "libraries/tinymce-apex/content.css",
            skin: false,
            content_style: CONTENT_STYLE,
            style_formats: [
                { title: "Paragraph", format: "p" },
                { title: "Heading 1", format: "h1" },
                { title: "Heading 2", format: "h2" },
                { title: "Heading 3", format: "h3" }
            ],
            text_patterns: [
                { start: "*", end: "*", format: "italic" },
                { start: "_", end: "_", format: "italic" },
                { start: "__", end: "__", format: "bold" },
                { start: "**", end: "**", format: "bold" },
                // not currently supported by our backend
                //{ start: "~~", end: "~~", format: "strikethrough"},
                { start: "`", end: "`", format: "code" },
                // does not work as expected
                //{ start: "> ", cmd: "mceBlockQuote" },
                { start: "#", format: "h1" },
                { start: "##", format: "h2" },
                { start: "###", format: "h3" },
                { start: "####", format: "h4" },
                { start: "#####", format: "h5" },
                { start: "######", format: "h6" },
                // The following text patterns require the `lists` plugin
                { start: "1. ", cmd: "InsertOrderedList" },
                { start: "* ", cmd: "InsertUnorderedList" },
                { start: "- ", cmd: "InsertUnorderedList" },
                { start: "--- ", cmd: "InsertHorizontalRule"}
                // does not work as expected
                // also see https://github.com/tinymce/tinymce/issues/7854
                // { start: "````", replacement: "<pre><code></code></pre>" }
            ],
            codesample_global_prismjs: true,
            codesample_languages: [
                { text: "Plaintext", value: "plaintext" },
                { text: "HTML", value: "html" },
                { text: "CSS", value: "css" },
                { text: "JavaScript", value: "js" },
                { text: "SQL", value: "sql" }
            ]
        };

        if ( mode === MODE_HTML ) {
            if ( toolbarVariant === TOOLBAR_BASIC ) {
                options.toolbar = "bold italic bullist numlist undo redo";
            } else if ( toolbarVariant === TOOLBAR_INTERMEDIATE ) {
                options.toolbar = "styles bold italic underline strikethrough link bullist numlist blockquote table undo redo";
            } else if ( toolbarVariant === TOOLBAR_FULL ) {
                // for full, we split it into 3 individual, labeled toolbars
                // note toolbar labeling requires items to be passed as an array
                options.toolbar =  [
                    { name: MSG_TOOLBAR_FORMATTING, items: "styles bold italic underline strikethrough subscript superscript removeformat".split( " " ) },
                    { name: MSG_TOOLBAR_ALIGNMENT, items: "alignjustify alignleft aligncenter alignright indent outdent".split( " " ) },
                    { name: MSG_TOOLBAR_LISTS, items: "bullist numlist".split( " " ) },
                    { name: MSG_TOOLBAR_FONT, items: "fontsize fontfamily forecolor backcolor".split( " " ) },
                    { name: MSG_TOOLBAR_EXTRAS, items: "link blockquote pagebreak hr table codeSample preview code undo redo".split( " " ) }
                ];
            }
            if ( toolbarStyle === TOOLBAR_STYLE_OVERFLOW ) {
                options.toolbar_mode = "sliding";
            } else if ( toolbarStyle === TOOLBAR_STYLE_MULTILINE ) {
                options.toolbar_mode = "wrap";
            }
            options.image_caption = true;
        } else if ( mode === MODE_MARKDOWN ) {
            options.toolbar_style = "wrap";
            options.toolbar = "styles bold italic inline-code link bullist numlist blockquote codesample undo redo";
            // allowed html elements and attributes
            // tinymce bug: data attributes ( or attributes with hyphen? ) are not removed
            // https://github.com/tinymce/tinymce/issues/8046
            // should be kept in sync with our backend parser
            options.valid_elements = [
                "img[src|alt|title]",
                "br",
                "a[href|title|target]",
                "em",
                "strong",
                "p",
                "h1", "h2", "h3", "h4", "h5", "h6",
                "pre[class]",
                "code[class]",
                "hr",
                "blockquote",
                "ul",
                "ol",
                "li",
                "em/i",         // convert i to em
                "strong/bold"   // convert bold to strong
            ].join( "," );
        }

        return options;
    };

    const markdown2html = markdown => {
        return util.sanitizeHtml( window.marked.parse( markdown, {
            gfm: true,
            breaks: true,
            tables: true,
            mangle: false,
            xhtml: false,
            headerIds: false
        } ) );
    };

    const html2markdown = html => {
        return turndownService.turndown( html );
    };

    /**
     * See 35207213
     * The TinyMCE Code Sample plug-in mishandles code blocks
     * It outputs them as:
     *  <pre class="language-..."><code>...</code></pre>
     * As opposed to:
     *  <pre><code class="language-...">...</code></pre>
     *
     * Further, if the markup does not match the first variant,
     *  the element won't be considered a code block.
     *
     * All of our other Markdown components expect the language class on the code element
     *  - apex_markdown
     *  - markedjs
     *  - turndownjs
     *  - ckeditor5
     *
     * TinyMCE is the only component that does it differently.
     *
     * Ideally, we need to rewrite the TinyMCE Code Sample plug-in to fix this behavior.
     * But until then, we will manually pre, and post-process the content to fit our needs.
     *
     * On setValue, we need to move the class from code to pre to make the plug-in happy
     * On getValue, we need to move the class from pre to code to have correct output
     *
     * We do this via 2 helper plug-ins, "codesample-pre" and "codesample-post"
     * "codesample-pre" must run before "codesample" and "codesample-post" after
     *
     * See the setup() call in images/libraries/tinymce/x.x.x/plugins/codesample/plugin.js
     * This code complements that code.
     */
    tinymce.PluginManager.add( "codesample-pre", editor => {
        editor.on( "SetContent", () => {
            const dom = editor.dom;
            const codeElems = dom.select( "code" );
            codeElems.forEach( codeElem => {
                const parentElement = codeElem.parentElement;
                if ( parentElement && parentElement.nodeName === "PRE" ) {
                    codeElem.classList.forEach( cls => {
                        parentElement.classList.add( cls );
                        codeElem.classList.remove( cls );
                    } );
                    if ( !parentElement.classList.length ) {
                        parentElement.classList.add( "language-plaintext" );
                    }
                }
            } );
        } );

        editor.on( "PreInit", () => {
            editor.parser.addNodeFilter( "pre", nodes => {
                nodes.forEach( node => {
                    if ( node.children().length && node.firstChild.name === "code") {
                        node.attr( "contenteditable", "false" );
                        node.attr( "data-mce-highlighted", "false" );
                    }
                } );
            } );
        } );
    } );

    tinymce.PluginManager.add( "codesample-post", editor => {
        editor.on( "PreProcess", e => {
            const dom = editor.dom;
            const codeElems = dom.select( "code", e.node );
            codeElems.forEach( codeElem => {
                const parentElement = codeElem.parentElement;
                if ( parentElement && parentElement.nodeName === "PRE" ) {
                    parentElement.classList.forEach( cls => {
                        if ( cls.startsWith( "language-" ) ) {
                            if ( cls !== "language-plaintext" ) {
                                codeElem.classList.add( cls );
                            }
                            parentElement.classList.remove( cls );
                        }
                    } );
                }
            } );
        } );
    } );

    // note that for now, none of these attributes are reactive
    // tinymce likewise cannot be changed via these attributes
    //  that should to be done via the Init Js Function
    const componentMetaData = {
        // markdown or html, defaults to markdown
        mode: {
            type: WebComponent.dataTypes.enum,
            defaultValue: MODE_MARKDOWN,
            enumValues: [ MODE_MARKDOWN, MODE_HTML ]
        },
        // pass in the label of the item
        // to be used for accessibility purposes
        label: {
            type: WebComponent.dataTypes.string
        },
        // toolbar is available for html only
        toolbar: {
            type: WebComponent.dataTypes.string
        },
        // toolbar style is available for html only
        toolbarStyle: {
            type: WebComponent.dataTypes.string
        },
        // the mininum height of the editor
        // note that the min and max height include the toolbar
        minHeight: {
            type: WebComponent.dataTypes.number,
            defaultValue: DEFAULT_MIN_HEIGHT
        },
        // the maximum height of the editor
        // not specifying this option will make the editor grow infinitely
        maxHeight: {
            type: WebComponent.dataTypes.number
        },
        // how to return the value when displayValueFor is called
        //  for example, in an IG's non-editable view
        displayValueMode: {
            type: WebComponent.dataTypes.string,
            enumValues: [ DISPLAY_VALUE_RICH_TEXT, DISPLAY_VALUE_ESCAPED, DISPLAY_VALUE_PLAIN_TEXT ],
            defaultValue: DISPLAY_VALUE_PLAIN_TEXT
        },
        // the tinymce language. see the tinymce docs for more info
        // defaults to English
        // todo. consider the implications of moving or replicating
        //  the backend get_tiny_lang conversion to the front-end
        language: {
            type: WebComponent.dataTypes.string
        },
        // readonly flag
        readOnly: {
            type: WebComponent.dataTypes.boolean
        },
        // optional placeholder text
        placeholder: {
            type: WebComponent.dataTypes.string
        },
        // can configure the editor to display as a div showing the content + a button
        //  that when clicked opens the editor in a dialog. useful in the IG
        visualMode: {
            type: WebComponent.dataTypes.enum,
            defaultValue: VISUAL_MODE_INLINE,
            enumValues:  [ VISUAL_MODE_INLINE, VISUAL_MODE_DIALOG ]
        }
    };

    // note that the delegated attributes are always reactive
    const delegatedAttributes = [ "value", "required", "name" ];
    // merge them with the componentMetaData
    apex.item.ItemComponent._addCommonAttributes( delegatedAttributes, componentMetaData );
    // extend each object with attrName property -> the object key in hyphen-case
    WebComponent._addHyphenCaseAttrNames( componentMetaData );
    // get the attrNames of the objects
    const observedAttributes = WebComponent._configObservedAttributes( componentMetaData, delegatedAttributes );

    class RichTextEditor extends item.ItemComponent {

        // Properties

        // hidden input holding the value
        _wrappedElement$;

        #dummyElement$;     // passed as selector to tinymce
        #previewElement$;   // visual mode: dialog
        #openEditorButton$; // visual mode: dialog
        #dialog$;           // visual mode: dialog
        #placeholder$;

        #options;
        #editor;
        #initialValue;      // used to compute isChanged()
        #deferredObject$;

        #v = {};

        constructor() {
            super();
            this._configPrivateState( this.#v );
            // type hidden as opposed to a text field as it needs to preserve all characters, such as new lines
            this._wrappedElement$ = $( "<input type='hidden' class='u-hidden'/>" );
            this.#deferredObject$ = this._initDelayLoading();
        }

        _render( attrChanges ) {

            trace( "_render", attrChanges );

            const self = this;

            // in this iteration of the widget, there are no reactive attributes
            // and we don't support moving the editor around the DOMPurify
            // once it's rendered, it's rendered
            if ( this.rendered ) {
                return;
            }

            if ( self.#v.mode === MODE_MARKDOWN && !markdownDependenciesConfigured ) {
                configureMarkdownDependencies();
                markdownDependenciesConfigured = true;
            }

            self.#placeholder$ = self.element.find( "." + CLS_PLACEHOLDER );

            // staring with the dummy element hidden has the effect of hiding the tinymce editor as well
            // which helps avoid initial load jittering- which is caused primarily by the resizing plugin
            self.#dummyElement$ =
                $( "<div/>" )
                    .attr( "class", CLS_DUMMY )
                    .attr( "data-editor", self.id )
                    .hide();

            self.element.append( self._wrappedElement$ );

            if ( self.#v.visualMode === VISUAL_MODE_INLINE ) {
                self.element.append( self.#dummyElement$ );
            } else if ( self.#v.visualMode === VISUAL_MODE_DIALOG ) {

                self.element.append( $( `
                    <div class="apex-item-group apex-item-group--rich-text-editor">
                        <div class="${ CLS_PREVIEW + " " + ( self.#v.displayValueMode === DISPLAY_VALUE_RICH_TEXT ? CLS_PREVIEW_RICH_TEXT : "" )  }"
                             tabindex="0" role="textbox" aria-haspopup="dialog" aria-expanded="false">
                        </div>
                        <button class="${ CLS_BUTTON } a-Button a-Button--rte" aria-hidden="true" tabindex="-1" type="button">
                            <span class="a-Icon icon-ig-expand"></span>
                        </button>
                    </div>
                ` ) );

                self.#previewElement$ = self.element.find( "." + CLS_PREVIEW );
                self.#openEditorButton$ = self.element.find( "." + CLS_BUTTON );

                self.#openEditorButton$.on( "click" , function() {
                    self.#openDialog();
                } );

                self.#previewElement$.on( "keydown", function( e ) {
                    if ( e.which === keyCodes.SPACE ) {
                        self.#openDialog();
                    }
                } );

                const dialog$ = $( "<div/>" );

                dialog$.append( self.#dummyElement$ );

                dialog$.dialog( {
                    title: self.#v.label,
                    closeText: lang.getMessage( "APEX.DIALOG.CLOSE" ),
                    width: 640,
                    height: "auto",
                    autoOpen: false,
                    modal: true,
                    // we do want to close on escape, but we implement this ourselves
                    // as this might interfere with the toolbar Escaping mechanisms
                    closeOnEscape: false,
                    dialogClass: CLS_DIALOG,
                    buttons: [ {
                        text: lang.getMessage( "APEX.DIALOG.CANCEL" ),
                        click: function () {
                            dialog$.dialog( "close" );
                        }
                    }, {
                        text: lang.getMessage( "APEX.DIALOG.OK" ),
                        class: "js-confirmBtn ui-button--hot",
                        click: function () {
                            $( this ).data( "shouldSave", true );
                            dialog$.dialog( "close" );
                        }
                    } ],
                    open: function() {
                        $( this ).data( "shouldSave", false );
                    },
                    close: function() {
                        if ( $( this ).data( "shouldSave" ) ) {
                            // commits the value on OK
                            self.#commitTransientValue();
                        } else {
                            // resets the value on escape, cancel, close
                            self.#setTransientValue( self.getValue() );
                        }
                    }
                } );

                self.#dialog$ = dialog$;
            }

            // editor options
            let editorOptions = getDefaultEditorOptions( self.#v.mode, self.#v.toolbar, self.#v.toolbarStyle );

            editorOptions.language = self.#v.language;
            editorOptions.readonly = self.#v.readOnly;
            editorOptions.placeholder = self.#v.placeholder;
            editorOptions.min_height = self.#v.minHeight;
            editorOptions.max_height = self.#v.maxHeight;
            // despite the name, this option actually sets the *aria-label* of the contenteditable *body element* of the iframe
            // the iframe will continue to have the *title* of "Rich Text Area"
            // known bug: https://github.com/tinymce/tinymce/issues/8310
            // however this bug/ undocumented behavior works fine for us
            editorOptions.iframe_aria_text = self.#v.label;

            // todo. remove these two options when uptaking v7.0.0
            editorOptions.sandbox_iframes = true;
            editorOptions.convert_unsafe_embeds = true;

            const options = {
                editorOptions: editorOptions
            };

            self.#options = self._applyInitFunc( options );

            // if the editor exists inside an inline dialog
            // we should allow the initialization of the dialog to take place first
            // otherwise the iframe would be moved around in the dom, causing issues on firefox. see #36121870
            // thus the setTimeout
            setTimeout( () => {
                window.tinymce.init( {
                    selector: `.${ CLS_DUMMY }[data-editor="${ self.id }"]`,
                    ...( self.#options.editorOptions ),
                    setup: function( editor ) {

                        // add our custom inline <code> button
                        editor.ui.registry.addToggleButton( "inline-code", {
                            icon: "sourcecode",
                            tooltip: "code",
                            onAction: () => {
                                if ( !editor.formatter.match( "code" ) ) {
                                    editor.execCommand( "RemoveFormat" );
                                }
                                editor.execCommand( "mceToggleFormat", false, "code" );
                            },
                            onSetup: ( buttonApi ) => {
                                const editorEventCallback = ( eventApi ) => {
                                    let enabled = true;
                                    let active = false;
                                    eventApi.parents.forEach( parent => {
                                        if ( parent.tagName === "PRE" ) {
                                            enabled = false;
                                        }
                                    } );
                                    if ( enabled && editor.formatter.match( "code" ) ) {
                                        active = true;
                                    }
                                    buttonApi.setEnabled( enabled );
                                    buttonApi.setActive( active );
                                };

                                editor.on( "NodeChange", editorEventCallback );

                                /* onSetup should always return the unbind handlers */
                                return () => {
                                    editor.off( "NodeChange", editorEventCallback );
                                };
                            }
                        } );

                        editor.on( "init", () => {

                            self.#editor = editor;

                            // change-event logic for inline mode
                            // we don't wish to trigger a change event on each keystroke
                            // rather, we store a temporary value on focusin, and only trigger
                            // a change event on focusout if the value has really changed
                            if ( self.#v.visualMode === VISUAL_MODE_INLINE ) {
                                let tempValue;
                                editor.on( "focus", () => {
                                    tempValue = self.#getTransientValue();
                                } );

                                editor.on( "blur", () => {
                                    const newValue = self.#getTransientValue();
                                    if ( tempValue !== newValue ){
                                        self.#commitTransientValue();
                                    }
                                } );

                                // todo. blur the editor as the window loses focus
                                // editor.fire( "blur" )
                            }

                            // dialog specific logic
                            if ( self.#v.visualMode === VISUAL_MODE_DIALOG ) {
                                // escape should reset the value and close
                                editor.on( "keydown", e => {
                                    if ( e.which === keyCodes.ESCAPE ) {
                                        self.#dialog$.dialog( "close" );
                                    }
                                } );

                                // on editor resize, resize the dialog
                                editor.on( "ResizeEditor", () => {
                                    self.#dialog$.dialog( "option", "height", "auto" );
                                } );

                                // needed by the IG integration
                                // as the editor is focused, because we're working with an iframe,
                                // the grid will wrongly assume focus has been lost, and it will _beginDeactivate
                                // we must therefore _clearDeactivate, by re-triggering a focusin on the iframe.
                                // if we don't do this, focus won't be returned to the cell on dialog close.
                                // note that we must use focusin and not focus
                                // see #35314916
                                editor.on( "focusin", () => {
                                    $( editor.iframeElement ).trigger( "focusin" );
                                } );
                            }

                            // add focused class as appropriate. used for focus styling
                            const iframe$ = $( editor.iframeElement ),
                                body$ = $( editor.getBody() );

                            editor.on( "focus", () => {
                                iframe$.addClass( CLS_IS_FOCUSED );
                                body$.addClass( CLS_IS_FOCUSED );

                                // on focus ensure that editor is scrolled into viewport. see 35120575
                                const bounding = editor.iframeElement.getBoundingClientRect();

                                if ( !( bounding.top <= ( window.innerHeight || document.documentElement.clientHeight ) &&
                                        bounding.left >= 0 &&
                                        bounding.right <= ( window.innerWidth || document.documentElement.clientWidth ) &&
                                        bounding.bottom >= theme.defaultStickyTop() )
                                ){
                                    self.element[0].scrollIntoView(false);
                                }
                            } );

                            editor.on( "blur", () => {
                                iframe$.removeClass( CLS_IS_FOCUSED );
                                body$.removeClass( CLS_IS_FOCUSED );
                            } );

                            // since the dummy element is initially hidden, the editor will be initialized hidden as well
                            // showing it again here, once everything is loaded
                            // note that setting the editor value should be done after it's shown due to a tinymce restriction
                            editor.show();

                            // the #initialValue is used in isChanged
                            // the editor might choose to format or slightly alter the content on load,
                            // but this shouldn't be considered a change, so we set the initivalValue when all is ready
                            // using one initial getValue
                            self.setValue( self.#v.value, null, true );
                            self.#initialValue = self.getValue();

                            self.#placeholder$.remove();

                            self.#applyA11yFixes();

                            self.#deferredObject$.resolve();
                        } );
                    }
                } );
            }, 0 );
        }

        getValue() {
            trace( "getValue" );
            return this._wrappedElement$[0].value;
        }

        setValue( value, displayValue, suppressChangeEvent ) {
            trace( "setValue", value, displayValue, suppressChangeEvent );

            // gracefully ignore the automatic pre _render setValue call
            // we don't need it. relying on #v.value inside of _render instead
            if ( !this.rendered ) {
                trace( "ignoring setValue" );
                return;
            }

            // the converters don't play well with non-string values
            // this is the one point where we fix such values
            value ??= "";

            if ( typeof value !== "string" ) {
                value = "" + value;
            }

            // content previously created by CKE5 might have been wrapped in a div.ck-content
            // for tinymce, this wrapper is no longer needed. see 35207154
            if ( value.length && value.startsWith( "<div class=\"ck-content\">" ) && value.endsWith( "</div>" ) ) {
                value = value.slice( 24, -6 );
            }

            this.#setTransientValue( value );
            this.#commitTransientValue( !suppressChangeEvent );
        }

        enable() {
            this.#editor?.mode.set( TINY_ENABLED );
        }

        disable() {
            this.#editor?.mode.set( TINY_DISABLED );
        }

        isDisabled() {
            return this.#editor?.mode.get() === TINY_DISABLED;
        }

        displayValueFor( value ) {
            if ( this.#v.mode === MODE_HTML){
                if ( this.#v.displayValueMode === DISPLAY_VALUE_RICH_TEXT ) {
                    return util.sanitizeHtml( value );
                } else if ( this.#v.displayValueMode === DISPLAY_VALUE_ESCAPED ) {
                    return apex.util.escapeHTML( value );
                } else if ( this.#v.displayValueMode === DISPLAY_VALUE_PLAIN_TEXT ) {
                    // replace new line elements with a space, strip html tags, collapse consecutive whitespaces
                    return util.stripHTML( value.replace( /<p>|<p\/>|<br>|<br\/>|<br \/>|<BR>|\n|&nbsp;/g, " " ) ).replace(/\s\s+/g, " ");
                }
            } else if (this.#v.mode === MODE_MARKDOWN) {
                if ( this.#v.displayValueMode === DISPLAY_VALUE_RICH_TEXT ) {
                    return `<div class="is-markdownified">${ markdown2html( value ) }</div>`;
                } else if ( this.#v.displayValueMode === DISPLAY_VALUE_ESCAPED ) {
                    return util.escapeHTML( value );
                } else if ( this.#v.displayValueMode === DISPLAY_VALUE_PLAIN_TEXT ) {
                    // replace new line elements with a space, strip html tags, collapse consecutive whitespaces
                    return util.stripHTML( markdown2html( value ) ).replace( /<p>|<p\/>|<br>|<br\/>|<br \/>|<BR>|\n|&nbsp;/g, " " ).replace(/\s\s+/g, " ");
                }
            }
        }

        setFocusTo() {
            let self = this;
            return {
                focus: function(){
                    if ( self.#v.visualMode === VISUAL_MODE_INLINE ) {
                        self.#editor?.focus();
                    } else if ( self.#v.visualMode === VISUAL_MODE_DIALOG ) {
                        if ( self.#dialog$.dialog( "isOpen" ) ) {
                            self.#editor?.focus();
                        } else {
                            self.#previewElement$.focus();
                        }
                    }
                }
            };
        }

        isChanged() {
            // if the editor is not done loading, the content could not have changed
            if ( !this.#editor ) {
                return false;
            } else {
                return this.#initialValue !== this.getValue();
            }
        }

        getEditor() {
            return this.#editor;
        }

        getValidity() {
            const self = this;
            const result = {
                valueMissing: self.required && self.getValue().trim().length === 0
                /* tooLong:  */
                /* tooShort: */
            };

            result.valid = !Object.values( result ).some( value => value );

            return result;
        }

        getPopupSelector() {
            // when editing in the Interactive Grid
            // the editor would be hidden when various editor popups would receive focus
            // e.g. the Link popup
            // providing a general selector that all popups share fixes this
            return POPUP_SELECTOR;
        }

        reinit( value, display ) {

            trace( "reinit", value, display );

            this.setValue( value, null, true );
            this.#initialValue = this.getValue();
            this.#editor.undoManager.clear();
        }

        /*
         * Private methods
         */

        // gets the current tinymce value, not the actual wrappedElement value
        #getTransientValue() {
            const rawContent = this.#editor.getContent();

            if ( this.#v.mode === MODE_HTML ) {
                return util.sanitizeHtml( rawContent );
            } else if ( this.#v.mode === MODE_MARKDOWN ) {
                return html2markdown( rawContent );
            }
        }

        // sets the current tinymce value, not the actual wrappedElement value
        #setTransientValue( value ) {
            if ( this.#v.mode === MODE_HTML ) {
                value = util.sanitizeHtml( value );
            } else if ( this.#v.mode === MODE_MARKDOWN ) {
                // sanitizes internally
                value = markdown2html( value );
            }

            this.#editor.setContent( value );
        }

        // commits the transient value of the editor to the wrappedElement
        #commitTransientValue( triggerChangeEvent = true ) {

            const finalValue = this.#getTransientValue();

            this._wrappedElement$.val( finalValue );

            if ( this.#previewElement$ ) {
                trace( "setting value of #previewElement$" );
                this.#previewElement$.html( this.displayValueFor( finalValue ) );
            }

            if ( triggerChangeEvent ) {
                this.element.trigger( "change" );
            }
        }

        #openDialog() {
            this.#setCursorToEnd();
            this.#dialog$.dialog( "open" );
            this.#editor.focus();
        }

        #setCursorToEnd() {
            this.#editor.selection.select( this.#editor.getBody(), true );
            this.#editor.selection.collapse( false );
        }

        // applies various accessibility fixes
        #applyA11yFixes() {
            const
                containerElem = this.#editor.container,
                iframeElem = this.#editor.iframeElement,
                bodyElem = this.#editor.getBody();

            // the container element, which already has role="application" should get an aria-label
            containerElem.setAttribute( "aria-label", MSG_RICH_TEXT_EDITOR );

            // the iframe element should not have a title, and role="none"
            iframeElem.removeAttribute( "title" );
            iframeElem.setAttribute( "role", "none" );

            // the body element should have a role="textbox" and aria-multiline="true"
            bodyElem.setAttribute( "role", "textbox" );
            bodyElem.setAttribute( "aria-multiline", "true" );

            if ( this.#v.readOnly || this.isDisabled() ) {
                bodyElem.setAttribute( "aria-roledescription", MSG_READ_ONLY_RICH_TEXT_EDITOR );
                bodyElem.setAttribute( "aria-readonly", "true" );
            } else {
                bodyElem.setAttribute( "aria-roledescription", MSG_RICH_TEXT_EDITOR );
                bodyElem.setAttribute( "aria-readonly", "false" );
            }

            // remove misused complementary role from unused, bogus element
            // https://github.com/tinymce/tinymce/issues/7639
            $( this.element ).find( ".tox-sidebar" ).removeAttr( "role" ).hide();
        }

        /*
         * Static methods
         */
        static get observedAttributes() {
            return observedAttributes;
        }
        static get delegatedAttributes() {
            return delegatedAttributes;
        }
        static get componentMetaData() {
            return componentMetaData;
        }
    }

    customElements.define( "a-rich-text-editor", RichTextEditor );

    /*
     * Allow various editor popups to work in jQuery UI dialogs
     */
    $.widget( "ui.dialog", $.ui.dialog, {
        _allowInteraction: function( event ) {
            return $( event.target ).closest( POPUP_SELECTOR ).length > 0 || this._super( event );
        }
    } );

    // there is a jQuery UI bug due to which dialogs lose focus when dragged
    // in an RTE + IG context, this would mean that when dragging the RTE dialog, the IG cell would get deselected
    //  causing usability and accessibility issues
    // the initial issue was fixed in https://github.com/jquery/jquery-ui/commit/1a696678878f48748912b4c9d5df6da91c760b98 (commit 1)
    // however a regression was introduced in https://github.com/jquery/jquery-ui/commit/60486ac632a0a1bbbb0c7449fe17bccfae11af80 (commit 2)
    // attempting here to fix the code in commit 2
    $.widget( "ui.dialog", $.ui.dialog, {
        _create: function() {
            this._super();
            // although other parts of APEX could benefit from this fix
            // there is currently no place to apply it globally. the jQueryUI overrides in core.js
            //  kick in after the RTE's dialog has been created, which is too late
            // once there is a better place to add such overrides globally, this code should be moved
            // and the restriction on CLS_DIALOG should be dropped
            if ( $( this ).closest( `.ui-dialog.${ CLS_DIALOG }` ) ) {
                const uiDialog = this.uiDialog;
                if ( this.uiDialogTitlebar.length ) {
                    this.uiDialogTitlebar.on( "mousedown", function( event ) {
                        // fixing error from commit 2. adding the missing ".length"   vvvvvvv
                        if ( !$( event.target ).closest( ".ui-dialog-titlebar-close" ).length ) {
                            uiDialog.trigger( "focus" );
                        }
                    } );
                }
            }
        }
    } );

} )( apex.item, apex.util, apex.debug, apex.jQuery, apex.WebComponent, apex.lang, apex.theme );
