/*!
 * Copyright (c) 2023, 2024, Oracle and/or its affiliates.
 */

/* eslint quotes: ["error", "double"] */
/* eslint no-var: "error" */
/* eslint "comma-dangle": ["error", "never"] */
/* eslint keyword-spacing: ["error", { "before": true, "after": true } ] */
/* global DiagramBuilder */

( () => {
    "use strict";

    const { jQuery: $, debug, region, util, widget, WebComponent, actions, lang } = apex;
    const widgetUtil = widget.util;

    const DIAGRAM_CONFIG = {
        readOnly: true,
        renderStencil: false,
        showNavigator: false,
        selectionMode: "none",
        linkDesignations: [ {
            typeId: "ERROR",
            glyph: "\ue317"
        }, {
            typeId: "TIMEOUT",
            glyph: "\ue108"
        } ],
        style: {
            paperCls: "paper",
            toastForegroundContainerCls: "a-WorkflowToast"
        },
        routeHighlights: [ {
            typeId: "TERMINATED",
            color: "var(--a-diagram-route-terminated)",
            allowHighlightedOutboundLinks: false
        }, {
            typeId: "FAULTED",
            color: "var(--a-diagram-route-faulted)",
            allowHighlightedOutboundLinks: false
        }, {
            typeId: "ACTIVE",
            color: " var(--a-diagram-route-active)",
            allowHighlightedOutboundLinks: false
        }, {
            typeId: "COMPLETED",
            color: "var(--a-diagram-route-completed)"
        }, {
            typeId: "WAITING",
            color: "var(--a-diagram-route-waiting)",
            allowHighlightedOutboundLinks: false
        }, {
            typeId: "SUSPENDED",
            color: "var(--a-diagram-route-suspended)",
            allowHighlightedOutboundLinks: false
        } ]
    };

    const TOOLBAR_ACTIONS = {
        zoomIn: "zoom-in",
        zoomOut: "zoom-out",
        zoomReset: "zoom-reset",
        showNavigator: "show-navigator"
    };
    const ZOOM_STEP = .1;

    class WorkflowDiagram extends WebComponent {
        diagramElement;
        toolbarElement;
        targetElement;

        #diagram;
        #buttons;
        #initZoom;
        #workflowConfig;
        #context;

        #autoLoad;
        #regionId;
        #region$;
        #workflowIdItem;
        #ajaxIdentifier;

        #navigatorShown = false;

        #suspendedLoad = false;
        #loading = false;

        constructor() {
            super();
            const buttons = this.getAttribute( "buttons" );

            this.#ajaxIdentifier = this.getAttribute( "ajax-identifier" );
            this.#regionId = this.getAttribute( "region-id" );
            this.#region$ = $( `#${ util.escapeCSS( this.#regionId ) }`);
            this.#workflowIdItem = this.getAttribute( "workflow-id-item" );
            this.#initZoom = this.getAttribute( "init-zoom" );
            this.#autoLoad = this.hasAttribute( "auto-load" ) ? ( this.getAttribute( "auto-load" ).toLowerCase() !== "false" )  : true;
            this.#workflowConfig = JSON.parse( ( this.getAttribute( "workflow-config" ) || "[]" ) );

            if ( !this.#region$.length ) {
                debug.error( "Workfow Diagram region must use a template with ID attribute" );
            }

            this.#buttons = buttons ? buttons.split( ":" ) : [];

            region.create( this.#regionId, {
                type: "WorkflowDiagram",
                refresh: this.refresh.bind( this )
            } );
        }

        _render(){
            if ( this.rendered ) {
                return;
            }

            const el = document.createElement( "div" );
            this.targetElement = el;
            el.classList.add( "a-WorkflowApp" );
            el.setAttribute( "role", "img" );
            el.setAttribute( "aria-label", lang.getMessage( "APEX.WF_DIAGRAM.ARIA_LABEL" ) );
            el.setAttribute( "aria-description", lang.getMessage( "APEX.WF_DIAGRAM.ARIA_DESC" ) );
            this.append( el );

            const KeyboardAction = DiagramBuilder.KeyboardAction;
            this.#diagram = new DiagramBuilder( {
                ...DIAGRAM_CONFIG,
                locale: {
                    [ DiagramBuilder.Locale.STR_ZOOM_LEVEL ]: lang.getMessage( "APEX.WF.ZOOM_LEVEL" )
                },
                keyboardNavigationMap: {
                    "up": KeyboardAction.PAPER_PAN_UP,
                    "down": KeyboardAction.PAPER_PAN_DOWN,
                    "left": KeyboardAction.PAPER_PAN_LEFT,
                    "right": KeyboardAction.PAPER_PAN_RIGHT,
                    "ctrl+0": KeyboardAction.ZOOM_RESET,
                    "ctrl+minus": KeyboardAction.ZOOM_OUT,
                    "ctrl+plus": KeyboardAction.ZOOM_IN
                },
                el,
                initialZoom: parseFloat( this.#initZoom )
            } );
            this.#initToolbar();
            this.#createContext();

            widgetUtil.onVisibilityChange( $( this ), show => {
                if ( show && this.#suspendedLoad ) {
                    this.#suspendedLoad = false;
                    this.refresh();
                }
            } );

            if ( this.#autoLoad ) {
                this.refresh();
            }
        }

        _fetchData() {
            return new Promise( ( resolve, reject ) => {
                this.#loading = true;
                apex.server.plugin( {
                    regions: [ {
                        id: this.#regionId,
                        ajaxIdentifier: this.#ajaxIdentifier
                    } ],
                    pageItems: this.#workflowIdItem ? [ this.#workflowIdItem ] : []
                }, {
                    refreshObject: this,
                    loadingIndicator: this,
                    loadingIndicatorPosition: "centered",
                    success: res => {
                        this.#loading = false;
                        resolve( res.regions[ 0 ].result );
                    },
                    error: ( xhr, textStatus, errorThrown ) => {
                        this.#loading = false;
                        reject( xhr, textStatus, errorThrown );
                    }
                } );
            } );
        }

        refresh() {
            if ( !this.#diagram || this.#loading ) {
                return;
            }

            if ( widgetUtil.isVisible( $( this ) ) ) {
                this._fetchData().then( data => {
                    const activities = [];
                    const links = [];
                    const route = [];
                    const notStatusCompletedHighlights = {};

                    data.forEach(entry => {
                        if ( activities.findIndex( act => act.id === entry.activityId ) < 0 ) {
                            activities.push( {
                                id: entry.activityId,
                                text: entry.activityName,
                                typeId: entry.activityType,
                                diagram: entry.activityDiagram
                            } );
                        }

                        if ( entry.state !== "INCOMPLETE" ) {
                            const inRoute = route.find( item => item.cellId === entry.activityId && item.previousId === entry.previousId );
                            // The incoming data contains every element (i.e. route-step) as many times as many outbound links it has, so e.g. an element
                            // which should be in the route but has 3 outbound links, will be included 3x in the data for that route-step.
                            // We need to keep it only once for every step, i.e. we need to discard all the other "copies" - we distinguish them
                            // by different linkIds.
                            const { linkId } = entry;
                            if ( !inRoute || inRoute.linkId === linkId ) {
                                route.push( {
                                    cellId: entry.activityId,
                                    previousId: entry.previousId,
                                    highlightTypeId: entry.state,
                                    linkId
                                } );
                            }
                            // if it exists in the route, change the highlight-type if it is not "COMPLETED"
                            if ( entry.state !== "COMPLETED" ) {
                                notStatusCompletedHighlights[ entry.activityId ] = entry.state;
                            }
                        } 

                        if ( entry.linkId && links.findIndex( link => link.id === entry.linkId ) < 0 ) {
                            links.push( {
                                id: entry.linkId,
                                name: entry.linkName,
                                source: entry.activityId,
                                target: entry.linkTargetId,
                                diagram: entry.linkDiagram,
                                type: entry.linkType
                            } );
                        }
                    } );

                    // If we have some "not completed" highlights, change the state of all the occurences of those cells.
                    Object.entries( notStatusCompletedHighlights ).forEach( ( [ key, value] ) => {
                        route.filter( obj => obj.cellId === key ).forEach( obj => obj.highlightTypeId = value ); 
                    } );

                    this.#diagram.setElementsConfig( this.#getWorkflowActivitiesConfig( activities ) );
                    this.#diagram.load( {
                        cells: [ ...this.#prepareActivities( activities ), ...this.#prepareLinks( links ) ],
                        route
                    } );
                } );
            } else {
                this.#suspendedLoad = true;
            }
        }

        get diagramBuilder() {
            return this.#diagram;
        }

        #createContext(){
            const diagram = this.#diagram;
            const me = this;
            this.#context = actions.createContext( "workflowDiagram", this );
            this.#context.add( [ {
                name: TOOLBAR_ACTIONS.zoomIn,
                action: function() {
                    diagram.zoom( ZOOM_STEP );
                }
            }, {
                name: TOOLBAR_ACTIONS.zoomOut,
                action: function() {
                    diagram.zoom( -ZOOM_STEP );
                }
            }, {
                name: TOOLBAR_ACTIONS.zoomReset,
                action: function() {
                    diagram.zoom( 1, true );
                }
            }, {
                name: TOOLBAR_ACTIONS.showNavigator,
                get: function() {
                    return me.#navigatorShown;
                },
                set: function( value ) {
                    value ? diagram.showNavigator() : diagram.hideNavigator();
                    me.#navigatorShown = value;
                }
            } ] );
        }

        #initToolbar() {
            let config = {
                actionsContext: this.#context,
                simple: true,
                data: []
            };
            let zoomControls = [];

            if ( this.#buttons.includes( "ZOOM_IN" ) ) {
                zoomControls.push( {
                    type: "BUTTON",
                    title: lang.getMessage( "APEX.WF.ZOOM_IN" ),
                    label: lang.getMessage( "APEX.WF.ZOOM_IN" ),
                    iconOnly: true,
                    icon: "icon-zoom-region",
                    action: TOOLBAR_ACTIONS.zoomIn
                } );
            }

            if ( this.#buttons.includes( "ZOOM_OUT" ) ) {
                zoomControls.push( {
                    type: "BUTTON",
                    title: lang.getMessage( "APEX.WF.ZOOM_OUT" ),
                    label: lang.getMessage( "APEX.WF.ZOOM_OUT" ),
                    iconOnly: true,
                    icon: "icon-zoom-page",
                    action: TOOLBAR_ACTIONS.zoomOut
                } );
            }

            if ( this.#buttons.includes( "ZOOM_RESET" ) ) {
                zoomControls.push( {
                    type: "BUTTON",
                    title: lang.getMessage( "APEX.WF.ZOOM_RESET" ),
                    label: lang.getMessage( "APEX.WF.ZOOM_RESET" ),
                    iconOnly: true,
                    icon: "icon-search",
                    action: TOOLBAR_ACTIONS.zoomReset
                } );
            }

            if ( zoomControls.length ) {
                config.data.push( {
                    id: "zoomControls",
                    align: "end",
                    groupTogether: true,
                    controls: zoomControls
                } );
            }

            if ( this.#buttons.includes( "NAVIGATOR" ) ) {
                config.data.push( {
                    id: "navigatorControls",
                    align: "end",
                    controls: [ {
                        type: "BUTTON",
                        title: lang.getMessage( "APEX.WF.NAVIGATOR" ),
                        label: lang.getMessage( "APEX.WF.NAVIGATOR" ),
                        iconOnly: true,
                        icon: "icon-navigator",
                        action: TOOLBAR_ACTIONS.showNavigator
                    } ]
                });
            }

            if ( config.data.length ) {
                const tEl = document.createElement( "div" );
                this.toolbarElement = tEl;
                tEl.classList.add( "a-Toolbar", "a-Toolbar-pageColumn", "a-Toolbar-pageColumn--gridLayout", "a-Toolbar--simple" );
                this.prepend( tEl );
                $( tEl ).toolbar( config );
            }
        }

        /**
         *
         * @param {array} links array of link objects
         * @returns {array}
         */
        #prepareLinks( links ) {
            return links.map( link => {
                let coordinates = JSON.parse( link.diagram );
                let sourceAnchor, targetAnchor, labels = [];
                if ( coordinates.source.args ) {
                    sourceAnchor = { args: coordinates.source.args, name: coordinates.source.name };
                }

                if ( coordinates.target.args ) {
                    targetAnchor = { args: coordinates.target.args, name: coordinates.target.name };
                }

                // hide the label of "normal" transitions
                if ( link.type !== "NORMAL" ) {
                    labels.push( {
                        text: link.name,
                        position: coordinates.label
                    } );
                }

                return {
                    id: link.id,
                    source: {
                        id: link.source,
                        anchor: sourceAnchor
                    },
                    target: {
                        id: link.target,
                        anchor: targetAnchor
                    },
                    vertices: coordinates.vertices,
                    z: coordinates.z,
                    designation: link.type,
                    labels,
                    ...( coordinates.router && { router: coordinates.router } )
                };
            } );
        }

        /**
         *
         * @param {array} activities array of activity objects
         * @returns {array}
         */
        #prepareActivities( activities ) {
            return activities.map( activity => {
                activity = {
                    ...activity,
                    ...JSON.parse( activity.diagram ),
                    hasMenu: false
                };
                delete activity.diagram;
                return activity;
            } );
        }

        /**
         *
         * @param {array} activities array of workflow activity objects
         * @returns {array} array of activity type objects
         */
        #getWorkflowActivitiesConfig( activities ) {
            let pluginType = this.#workflowConfig.filter( config => config.typeId === "$PLUGIN$")[ 0 ],
             types = this.#workflowConfig.reduce( ( acc, curr ) => {
                acc.push( curr.typeId );
                return acc;
            }, [] );
            return [ ...( activities.filter( activity => !types.includes( activity.typeId ) ).map( act => {
                return {
                    ...pluginType,
                    typeId: act.typeId
                } ;
            } ) ), ...this.#workflowConfig ];
        }
    }

    $( function() {
        customElements.define( "a-workflow-diagram", WorkflowDiagram );
    } );

} )();