/* @oracle/oraclejet-preact: 16.0.1 */
'use strict';

var textUtils = require('./textUtils-c1554e64.js');
require('./accUtils-c4465ee9.js');

/**
 * @license
 * Copyright (c) %FIRST_YEAR% %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
const AXIS_DEFAULTS = {
    titleGap: 6,
    labelGap: 6
};

/**
 * @license
 * Copyright (c) %FIRST_YEAR% %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
/**
 * Calculates the number of skips between axis labels and the dimensions of unskipped labels.
 * @param labels The labels to render.
 * @param getLabelDims The function that returns dimension of given label.
 * @param isHoriz Whether the axis is horizontal.
 * @param numOfSafeSkips The no of labels to skip safely.
 * @returns
 */
function skipLabelsUniform(labels, getLabelDims, isHoriz, numOfSafeSkips) {
    let numOfSkips = numOfSafeSkips || 0;
    let dimsA;
    let dimsB;
    const dims = [getLabelDims(labels[0], 0)];
    const isOverlapping = isHoriz ? textUtils.isWidthOverlap : textUtils.isHeightOverlap;
    let indexA = 0;
    let indexB = numOfSkips + 1;
    while (indexB < labels.length) {
        dimsA = dims[indexA] || getLabelDims(labels[indexA], indexA);
        if (!dims[indexA])
            dims[indexA] = dimsA;
        dimsB = dims[indexB] || getLabelDims(labels[indexB], indexB);
        if (!dims[indexB])
            dims[indexB] = dimsB;
        if (isOverlapping(dimsA, dimsB)) {
            numOfSkips += 1;
            indexA = 0;
            indexB = numOfSkips + 1;
        }
        else {
            indexA = indexB;
            indexB = indexA + numOfSkips + 1;
        }
    }
    return { numOfSkips, dims };
}
/**
 * Greedily skip labels. We will skip next label if it overlaps with the previous one.
 * This might not result in uniformly spaced labels.
 * @param labels
 * @param getLabelDims
 * @param isHoriz
 * @returns
 */
function skipLabelsGreedy(labels, getLabelDims, isHoriz) {
    const isOverlapping = isHoriz ? textUtils.isWidthOverlap : textUtils.isHeightOverlap;
    let currentDims = getLabelDims(labels[0], 0);
    const renderedLabels = [
        {
            label: labels[0],
            index: 0,
            labelProps: currentDims
        }
    ];
    let maxWidth = currentDims.width;
    let maxHeight = currentDims.height;
    for (let j = 1; j < labels.length; j++) {
        if (!labels[j])
            continue;
        const labelDim = getLabelDims(labels[j], j);
        if (!isOverlapping(currentDims, getLabelDims(labels[j], j))) {
            currentDims = labelDim;
            renderedLabels.push({
                label: labels[j],
                index: j,
                labelProps: labelDim
            });
            maxWidth = Math.max(maxWidth, labelDim.width);
            maxHeight = Math.max(maxHeight, labelDim.height);
        }
    }
    return { tickLabels: renderedLabels, maxWidth, maxHeight };
}
/**
 * Retuns the max dimension of the rendered labels after skipping overlapping labels.
 * @param labels An array of labels.
 * @param getLabelDims The function called to get dimension of the given label.
 * @param isHoriz If the overlap be measured in horizontal direction.
 * @param numOfSafeSkips Predetermined number labels to safely skip. This will optimize label
 * measurement as we don't have to measure all the labels.
 * @returns
 */
function getMaxTextDimensions(labels, getLabelDims, isHoriz, numOfSafeSkips) {
    const { numOfSkips, dims } = skipLabelsUniform(labels, getLabelDims, isHoriz, numOfSafeSkips);
    const tickLabels = [];
    let maxWidth = Number.MIN_VALUE;
    let maxHeight = Number.MIN_VALUE;
    for (let i = 0; i < labels.length; i += numOfSkips + 1) {
        maxWidth = Math.max(maxWidth, dims[i].width);
        maxHeight = Math.max(maxHeight, dims[i].height);
        tickLabels.push({
            index: i,
            label: labels[i],
            labelProps: dims[i]
        });
    }
    return {
        maxHeight,
        maxWidth,
        tickLabels,
        numOfSkips
    };
}
function updateLabelsDims(maxWidth, maxHeight, renderedLabelsInfo, getAxisPos, availSpace, position, isRtl) {
    renderedLabelsInfo.forEach((info) => {
        const dim = info.labelProps;
        let x;
        let y;
        if (position === 'left') {
            x = availSpace.x + maxWidth;
            y = getAxisPos(info.index);
        }
        else if (position === 'right') {
            x = availSpace.x + availSpace.width;
            y = getAxisPos(info.index);
        }
        else if (position === 'top') {
            x = getAxisPos(info.index) - (isRtl ? -1 : 1) * (dim.width / 2);
            y = availSpace.y + maxHeight - dim.height;
        }
        else {
            x = getAxisPos(info.index);
            y = availSpace.y + availSpace.height - maxHeight / 2 + AXIS_DEFAULTS.labelGap;
        }
        dim.x = x;
        dim.y = y;
    });
}
/**
 * Returns the enclosing rect for the label of given dimension.
 * @param dims The dimension
 * @param axisPosition The position of the axis.
 * @returns
 */
function getEnclosingRectDims(dims, axisPosition) {
    let x;
    let y;
    if (axisPosition === 'left') {
        x = dims.x - dims.width;
        y = dims.y - dims.height / 2;
    }
    else if (axisPosition === 'right') {
        x = dims.x;
        y = dims.y - dims.height / 2;
    }
    else {
        x = dims.x - dims.width / 2;
        y = dims.y - dims.height / 2;
    }
    const width = dims.width;
    const height = dims.height;
    return { x, y, width, height };
}
/**
 * Retuns the unaligned label dimensions that can be used by skipLabels.
 * @param coord
 * @param label
 * @param labelStyle
 * @param isHoriz
 * @param availSpace
 * @param getTextDimensions
 * @returns
 */
function getLabelDimensions(coord, label, labelStyle, isHoriz, availSpace, getTextDimensions) {
    if (!getTextDimensions) {
        return { width: 0, height: 0, x: 0, y: 0 };
    }
    const gap = AXIS_DEFAULTS.labelGap;
    const { width, height } = getTextDimensions(label, labelStyle);
    const labelWidth = isHoriz ? width + gap : width;
    const labelHeight = isHoriz ? height : height + gap;
    return {
        x: isHoriz ? coord - labelWidth / 2 : availSpace.x,
        width: labelWidth,
        height,
        y: isHoriz ? availSpace.y : coord + labelHeight / 2
    };
}

exports.AXIS_DEFAULTS = AXIS_DEFAULTS;
exports.getEnclosingRectDims = getEnclosingRectDims;
exports.getLabelDimensions = getLabelDimensions;
exports.getMaxTextDimensions = getMaxTextDimensions;
exports.skipLabelsGreedy = skipLabelsGreedy;
exports.updateLabelsDims = updateLabelsDims;
//# sourceMappingURL=labelUtils-47c1d5dd.js.map
