/* @oracle/oraclejet-preact: 16.0.1 */
import { jsx, Fragment, jsxs } from 'preact/jsx-runtime';
import { g as getGroupCenterCoord, f as findNearestDataPoint } from './layoutUtils-765327c6.js';
import { c as curveThroughPoints } from './PathUtils-36c1f409.js';
import { colorSchemeVars } from './Common/themes/themeContract.css.js';
import { r as rgb, a as rgba } from './utils-4aa3c8ad.js';
import { a as MARKER_PADDING, M as Marker } from './Marker-539b7f2e.js';
import './SvgShapesStyles.styles.css';
import { u as useTranslationBundle } from './useTranslationBundle-95f6039e.js';
import { g as getItemAriaLabel, e as getLineAreaChartNavUtil } from './navUtils-4a56ee79.js';
import { s as supportsMobileScreenReader } from './accUtils-7cc9e054.js';
import { C as Chart } from './Chart-61d61471.js';
import { u as useUser } from './useUser-6e653d84.js';
import { g as getColorRamp } from './colorUtils-89039001.js';
import { styles } from './UNSAFE_Chart/themes/ChartStyles.css.js';

/**
 * @license
 * Copyright (c) 2008 %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
/**
 * Returns a straight area path command.
 * @param {array} lineSegmentCoords Polyline top points array.
 * @param {array} bottomCoords Polyline bottom points array.
 * @return {string} Straight area path commands.
 */
function getStraightAreaPath(lineSegmentCoords, bottomCoords) {
    // Create the path data string
    let pathData;
    if (lineSegmentCoords.length > 0) {
        pathData =
            lineSegmentCoords
                .map((lineSegmentCoord, index) => {
                return `${index === 0 ? 'M' : 'L'} ${lineSegmentCoord[0]} ${lineSegmentCoord[1]}`;
            })
                .join(' ') +
                [...bottomCoords]
                    .reverse()
                    .map((bottomCoord) => {
                    return `L ${bottomCoord[0]} ${bottomCoord[1]}`;
                })
                    .join(' ') +
                'Z';
    }
    return pathData;
}
/**
 * Returns a curved area path command, based on cubic hermite splines, that goes through the points in the points array.
 * @param {array} lineSegmentCoords Polyline top points array.
 * @param {array} bottomCoords Polyline bottom points array.
 * @param {boolean} connectWithLine Whether the first point is reached using lineTo. Otherwise, moveTo is used.
 * @param {string} splineType The spline type.
 * @return {string} Curved area path commands.
 */
function getCurveAreaPath(lineSegmentCoords, bottomCoords, connectWithLine, splineType, isBottomSegmentCurved) {
    let pathData;
    if (lineSegmentCoords.length > 0) {
        if (!isBottomSegmentCurved) {
            pathData =
                curveThroughPoints(lineSegmentCoords, connectWithLine, splineType) +
                    [...bottomCoords]
                        .reverse()
                        .map((bottomCoord) => {
                        return `L ${bottomCoord[0]} ${bottomCoord[1]}`;
                    })
                        .join(' ') +
                    'Z';
        }
        else {
            pathData =
                curveThroughPoints(lineSegmentCoords, false, splineType) +
                    ' ' +
                    curveThroughPoints([...bottomCoords].reverse(), true, splineType) +
                    ' Z';
        }
    }
    return pathData;
}

/**
 * @license
 * Copyright (c) 2008 %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
function getCmdsForArea(lineSegmentCoords, bottomCoords, areaType, isHoriz, isBottomSegmentCurved) {
    if (areaType === 'straight') {
        return getStraightAreaPath(lineSegmentCoords, bottomCoords);
    }
    return getCurveAreaPath(lineSegmentCoords, bottomCoords, false, isHoriz ? 'mh' : 'mv', isBottomSegmentCurved);
}
function getCmdsForLine(lineSegmentCoords, lineType, isHoriz) {
    if (lineType === 'straight') {
        return lineSegmentCoords.join(' ');
    }
    return curveThroughPoints(lineSegmentCoords, false, isHoriz ? 'mh' : 'mv');
}
/**
 * Returns the cumulative series for stacked chart.
 */
function getCumulativeSeries(numSeries, numGroups, getDataItem, hiddenIds, isLog) {
    const series = [];
    for (let seriesIndex = 0; seriesIndex < numSeries; seriesIndex += 1) {
        const currentSeries = [];
        const prevSeries = series[seriesIndex - 1];
        for (let groupIndex = 0; groupIndex < numGroups; groupIndex += 1) {
            const item = getDataItem(seriesIndex, groupIndex);
            if (!item || hiddenIds?.has(item.id) || (isLog && item.value <= 0)) {
                continue;
            }
            if (seriesIndex) {
                currentSeries[groupIndex] = item.value + prevSeries[groupIndex];
            }
            else {
                currentSeries[groupIndex] = item.value;
            }
        }
        series.push(currentSeries);
    }
    return series;
}
/**
 * Returns the series data.
 */
function getSeriesData(numSeries, numGroups, getDataItem, hiddenIds, isLog) {
    const series = [];
    for (let seriesIndex = 0; seriesIndex < numSeries; seriesIndex += 1) {
        const currentSeries = [];
        for (let groupIndex = 0; groupIndex < numGroups; groupIndex += 1) {
            const item = getDataItem(seriesIndex, groupIndex);
            if (!item || hiddenIds?.has(item.id) || (isLog && item.value <= 0)) {
                continue;
            }
            currentSeries[groupIndex] = item.value;
        }
        series.push(currentSeries);
    }
    return series;
}
/**
 * Returns the markertype order.
 */
function getMarkers() {
    return ['square', 'circle', 'diamond', 'plus', 'triangleDown', 'triangleUp'];
}
/**
 * Returns the marker fill and stroke depending on the state.
 */
function getMarkerFillAndStroke(// TODO : JET-62244 - Replace getMarkerFillAndStroke function with classes.
isMarkerDisplayed, isSelectionEnabled, isHovered, isFocused, isSelected, seriesColor, itemColor) {
    let fill;
    let stroke;
    let outerStroke;
    if (isSelectionEnabled) {
        if (isSelected) {
            if (isMarkerDisplayed) {
                // TODO: settle on one visual for selected markers: 
                // Selected effect when markers are turned on
                fill = itemColor || seriesColor;
                outerStroke = rgb(colorSchemeVars.palette.neutral[130]);
                stroke = rgb(colorSchemeVars.palette.neutral[0]);
            }
            else {
                // Selected effect when markers are not turned on
                fill = rgb(colorSchemeVars.palette.neutral[0]);
                if (isHovered || isFocused) {
                    stroke = itemColor || seriesColor;
                }
                else {
                    stroke = rgb(colorSchemeVars.palette.neutral[130]);
                }
            }
        }
        else if (isHovered || isFocused) {
            stroke = itemColor || seriesColor;
            fill = rgb(colorSchemeVars.palette.neutral[0]);
        }
        else {
            if (isMarkerDisplayed) {
                fill = itemColor || seriesColor;
                stroke = rgb(colorSchemeVars.palette.neutral[0]);
            }
            else {
                fill = rgba(colorSchemeVars.palette.neutral[0], 0);
                stroke = rgba(colorSchemeVars.palette.neutral[0], 0);
            }
        }
    }
    else if (isMarkerDisplayed) {
        fill = itemColor || seriesColor;
        stroke = rgb(colorSchemeVars.palette.neutral[0]);
    }
    else {
        stroke = isFocused ? itemColor || seriesColor : rgba(colorSchemeVars.palette.neutral[0], 0);
        fill = rgba(colorSchemeVars.palette.neutral[0], 0);
    }
    return { fill, stroke, outerStroke };
}
/**
 * Returns the coordinates of line or area based on orientation and size of the chart.
 */
function getLineAreaSegments(yScale, xScale, data, startIndex, endIndex, isHorizontal, isLog, groups, seriesIndex, getDataItem, timeAxisType) {
    const segments = [];
    let currentPath = [];
    for (let groupIndex = startIndex; groupIndex < endIndex + 1; groupIndex += 1) {
        const item = data[groupIndex];
        if (!item || (isLog && item <= 0)) {
            segments.push(currentPath);
            currentPath = [];
            continue;
        }
        const x = isHorizontal
            ? yScale.transform(item)
            : getGroupCenterCoord(groups[groupIndex], groupIndex, xScale, getDataItem(seriesIndex, groupIndex), timeAxisType);
        const y = isHorizontal
            ? getGroupCenterCoord(groups[groupIndex], groupIndex, xScale, getDataItem(seriesIndex, groupIndex), timeAxisType)
            : yScale.transform(item);
        currentPath.push([x, y]);
        if (groupIndex === endIndex) {
            segments.push(currentPath);
        }
    }
    return segments;
}

function AreaSeries({ yScale, lineType = 'straight', color, isHorizontal, isLog, areaColorOpacity, lineSegmentCoords, bottomCoords, isHighlighted, isBottomSegmentCurved, seriesIndex }) {
    const style = {
        fill: color,
        opacity: (isHighlighted ? 1 : 0.3) * areaColorOpacity,
        pointerEvents: areaColorOpacity !== 1 ? 'none' : 'all'
    };
    return (jsx(Fragment, { children: lineSegmentCoords.map((lineSegmentCoords) => {
            let baseCoords;
            if (bottomCoords) {
                // custom bottom coords. eg: top of last series in stacked area
                baseCoords = bottomCoords;
            }
            else {
                // TODO: make this more generic when we support apps setting baseline
                // for log assume domain min is the baseline, otherwise use 0 baseline
                const baselineCoord = yScale.transform(isLog ? yScale.domain()[0] : 0);
                baseCoords = isHorizontal
                    ? [
                        [baselineCoord, lineSegmentCoords[0][1]],
                        [baselineCoord, lineSegmentCoords[lineSegmentCoords.length - 1][1]]
                    ]
                    : [
                        [lineSegmentCoords[0][0], baselineCoord],
                        [lineSegmentCoords[lineSegmentCoords.length - 1][0], baselineCoord]
                    ];
            }
            const cmds = getCmdsForArea(lineSegmentCoords, baseCoords, lineType, isHorizontal, isBottomSegmentCurved);
            return (jsx("path", { d: cmds, style: style, "data-oj-series-index": seriesIndex, "data-oj-object": 'area' }));
        }) }));
}

const MARKER_DEFAULT_SIZE = 10;
function Markers({ yScale, xScale, color, seriesIndex, startIndex, endIndex, hiddenIds, selectedIds, orientation, isLog, data, markerType, getDataItem, focusedItemIndex, hoveredItemIndex, isDrillEnabled, activeId, isSelectionEnabled, timeAxisType, series, groups }) {
    const isHoriz = orientation === 'horizontal';
    const items = [];
    const translations = useTranslationBundle('@oracle/oraclejet-preact');
    for (let groupIndex = startIndex; groupIndex < endIndex + 1; groupIndex++) {
        const item = getDataItem(seriesIndex, groupIndex);
        const value = data[groupIndex];
        if (!item || hiddenIds?.has(item.id) || (isLog && value <= 0)) {
            continue;
        }
        const x = isHoriz
            ? yScale.transform(value)
            : getGroupCenterCoord(groups[groupIndex], groupIndex, xScale, item, timeAxisType);
        const y = isHoriz
            ? getGroupCenterCoord(groups[groupIndex], groupIndex, xScale, item, timeAxisType)
            : yScale.transform(value);
        const isFocused = focusedItemIndex === groupIndex;
        const isHovered = hoveredItemIndex === groupIndex;
        const isSelected = !!selectedIds?.has(item.id);
        const isDrillable = item.drilling === 'on' || (item.drilling != 'off' && isDrillEnabled);
        const { fill, stroke, outerStroke } = getMarkerFillAndStroke(!!item.isMarkerDisplayed, isSelectionEnabled, isHovered, isFocused, isSelected, color, item.markerColor);
        const itemMarkerType = item.isMarkerDisplayed ? item.markerType || markerType : markerType;
        const isActive = isFocused || isHovered;
        const ariaLabel = supportsMobileScreenReader || isActive
            ? getItemAriaLabel(translations, series.name || series.id.toString(), groups[groupIndex].name || groups[groupIndex].id, item.value, isSelectionEnabled, isSelected, isDrillable, item.accessibleLabel)
            : undefined;
        const dataInfo = {
            'data-oj-series-index': seriesIndex,
            'data-oj-group-index': groupIndex,
            'data-oj-object': 'item',
            id: isActive ? activeId : undefined,
            role: ariaLabel ? 'img' : undefined,
            'aria-label': ariaLabel
        };
        const markerSize = (item.markerSize || MARKER_DEFAULT_SIZE) + 2 * MARKER_PADDING;
        items.push(jsx(Marker, { type: itemMarkerType, tx: x, ty: y, isInteractive: isDrillable || isSelectionEnabled, scale: 1, fill: fill, stroke: stroke, outerStroke: outerStroke, width: markerSize, height: markerSize, dataInfo: dataInfo, preserveRectAspectRatio: true }));
    }
    return jsx(Fragment, { children: items });
}

function LineSeries({ color, isHorizontal, lineType = 'straight', lineColor, lineStyle, lineWidth = 3, isHighlighted, lineSegmentCoords, seriesIndex }) {
    const style = {
        stroke: lineColor || color,
        strokeDasharray: lineStyle === 'dotted' ? 3 : lineStyle === 'dashed' ? 6 : undefined,
        strokeWidth: lineWidth,
        fill: rgba(colorSchemeVars.palette.neutral[0], 0)
    };
    return (jsx(Fragment, { children: lineSegmentCoords.map((value) => {
            const cmds = getCmdsForLine(value, lineType, isHorizontal);
            return lineType === 'straight' ? (jsx("polyline", { points: cmds, style: style, className: isHighlighted ? undefined : styles.dimmed, "data-oj-series-index": seriesIndex, "data-oj-object": 'line' })) : (jsx("path", { d: cmds, style: style, className: isHighlighted ? undefined : styles.dimmed, "data-oj-series-index": seriesIndex, "data-oj-object": 'line' }));
        }) }));
}

const AXIS_OFFSET = 0.2;
/**
 * The Line Area Chart displays information graphically using lines and filled areas, making relationships among the data easier to understand.
 */
function LineAreaChart({ width = '100%', height = '448px', selectedIds = [], hiddenIds = [], highlightedIds, hideAndShowBehavior = 'none', orientation = 'vertical', yAxis, series, groups, isStacked = false, getDataItem, selectionMode = 'none', drilling = 'off', type = 'line', ...props }) {
    const { direction } = useUser();
    const isRtl = direction === 'rtl';
    const isLog = yAxis?.scale === 'log';
    const hiddenSet = new Set(hiddenIds);
    const selectedSet = new Set(selectedIds);
    const highlightedSet = new Set(highlightedIds);
    const isHoriz = orientation === 'horizontal';
    const colors = getColorRamp();
    const renderedSeries = isStacked
        ? getCumulativeSeries(series.length, groups.length, getDataItem, hiddenSet, isLog)
        : getSeriesData(series?.length, groups?.length, getDataItem, hiddenSet, isLog);
    const renderedGetDataItem = (seriesIndex, groupIndex) => {
        if (!isStacked)
            return getDataItem(seriesIndex, groupIndex);
        return {
            ...getDataItem(seriesIndex, groupIndex),
            value: renderedSeries[seriesIndex][groupIndex]
        };
    };
    const getGapRatio = () => {
        return 1;
    };
    const getNextChartItemFunc = (startIndex, endIndex) => {
        return getLineAreaChartNavUtil(getDataItem, series.length, startIndex, endIndex, isStacked, isRtl, isHoriz);
    };
    const getDataItemPos = (xScale, yScale) => {
        return (seriesIndex, groupIndex) => {
            const value = renderedSeries[seriesIndex][groupIndex];
            const x = isHoriz ? yScale.transform(value) : xScale.transform(groupIndex);
            const y = isHoriz ? xScale.transform(groupIndex) : yScale.transform(value);
            return { x, y, height: 0, width: 0 };
        };
    };
    const findNearest = (xScale, yScale) => {
        const getSeriesDataInfo = () => {
            return {
                offset: 0,
                dataWidth: 0
            };
        };
        const getYCoord = (value, yScale) => {
            return yScale.transform(value);
        };
        return findNearestDataPoint(series, xScale, yScale, renderedGetDataItem, isStacked, getSeriesDataInfo, getYCoord, isLog, 0, hiddenSet);
    };
    const markers = getMarkers();
    const timeAxisType = props.xAxis?.timeAxisType;
    const AREA_OPACITY = type === 'lineWithArea' ? 0.5 : 1;
    const isAreaRendered = type === 'area' || type === 'lineWithArea';
    const isContrastLineNeeded = type === 'area';
    const allLineSegmentCoords = [];
    return (jsx(Chart, { width: width, height: height, selectedIds: selectedIds, hiddenIds: hiddenIds, hideAndShowBehavior: hideAndShowBehavior, orientation: orientation, yAxis: yAxis, series: series, groups: groups, selectionMode: selectionMode, drilling: drilling, isStacked: isStacked, getDataItem: getDataItem, findNearest: findNearest, getDataItemPos: getDataItemPos, renderGridLinesInFront: isAreaRendered, getNavUtil: getNextChartItemFunc, getGapRatio: getGapRatio, offset: AXIS_OFFSET, isRtl: isRtl, ...props, children: ({ xStartIndex, xEndIndex, xScale, yScale, activeId, focusedItemInfo, hoveredItemInfo }) => {
            return (jsxs(Fragment, { children: [renderedSeries.map((lineAreaSeries, index) => {
                        if (lineAreaSeries.length === 0) {
                            /* TODO: need a better way to handle when a series is hidden. */
                            return;
                        }
                        const seriesItem = series[index];
                        const lineSegmentCoords = getLineAreaSegments(yScale, xScale, lineAreaSeries, xStartIndex, xEndIndex, isHoriz, isLog, groups, index, renderedGetDataItem, timeAxisType);
                        allLineSegmentCoords[index] = lineSegmentCoords[0];
                        const isBottomSegmentCurved = seriesItem.lineType === 'curved' &&
                            isStacked &&
                            renderedSeries.length > 1 &&
                            index > 0;
                        const isHighlighted = highlightedIds === undefined || highlightedSet.has(seriesItem.id);
                        return (jsxs(Fragment, { children: [isAreaRendered && (jsx(AreaSeries, { yScale: yScale, lineType: seriesItem.lineType, color: seriesItem.areaColor || colors[index % colors.length], isHorizontal: isHoriz, areaColorOpacity: AREA_OPACITY, isBottomSegmentCurved: isBottomSegmentCurved, bottomCoords: isStacked && index != 0 ? allLineSegmentCoords[index - 1] : undefined, lineSegmentCoords: lineSegmentCoords, isLog: isLog, isHighlighted: isHighlighted, seriesIndex: index })), jsx(LineSeries, { lineType: seriesItem.lineType, 
                                    /* TODO: confirm with UX if this is contrastLine (changes depending on dark/light mode) or always white.
                                        It's always white in legacy.*/
                                    lineColor: isContrastLineNeeded ? colorSchemeVars.dvt.contrastLine : seriesItem.lineColor, 
                                    /* The area generated in legacy are bit apart and not touching when stacked. Preact areas are a bit closer than legacy when curved.
                                      Increased the contrast linewidth (to 2.5 from legacy 1.25) to give legacy visual treatment.
                                      TODO: investivage why legacy areas are not touching. maybe we want to do the same in preact svg utils that
                                      generates this curve.
                                     */
                                    lineWidth: isContrastLineNeeded ? (isStacked ? 2.5 : 1.25) : seriesItem.lineWidth, lineStyle: seriesItem.lineStyle, color: seriesItem.lineColor || colors[index % colors.length], lineSegmentCoords: lineSegmentCoords, isHorizontal: isHoriz, isHighlighted: isHighlighted, seriesIndex: index }), 
                                /*for non stacked charts, markers should render in the order of series.
                                 some markers might get hidden behind area but this preserves the legacy behavior. */
                                !isStacked && (jsx(Markers, { yScale: yScale, xScale: xScale, groups: groups, series: seriesItem, timeAxisType: timeAxisType, data: lineAreaSeries, color: seriesItem.lineColor ||
                                        seriesItem.areaColor ||
                                        colors[index % colors.length] /* TODO : make checks smarter. */, seriesIndex: index, startIndex: xStartIndex, endIndex: xEndIndex, markerType: markers[index % markers.length], orientation: orientation, isLog: isLog, activeId: activeId, getDataItem: getDataItem, selectedIds: selectedSet, focusedItemIndex: focusedItemInfo.seriesIndex === index && focusedItemInfo.isFocusVisible
                                        ? focusedItemInfo.groupIndex
                                        : undefined, hoveredItemIndex: hoveredItemInfo?.isCurrent && hoveredItemInfo.seriesIndex === index
                                        ? hoveredItemInfo.groupIndex
                                        : undefined, isSelectionEnabled: selectionMode === 'single' || selectionMode === 'multiple', isDrillEnabled: drilling === 'on' }))] }));
                    }), isStacked &&
                        renderedSeries.map((lineAreaSeries, index) => {
                            if (lineAreaSeries.length === 0) {
                                /* TODO: need a better way to handle when a series is hidden.*/
                                return;
                            }
                            const seriesItem = series[index];
                            return (jsx(Markers, { yScale: yScale, series: series[index], xScale: xScale, groups: groups, timeAxisType: timeAxisType, data: lineAreaSeries, color: seriesItem.lineColor || seriesItem.areaColor || colors[index % colors.length] // TODO : make checks smarter.
                                , seriesIndex: index, startIndex: xStartIndex, endIndex: xEndIndex, markerType: markers[index % markers.length], orientation: orientation, isLog: isLog, activeId: activeId, getDataItem: getDataItem, selectedIds: selectedSet, focusedItemIndex: focusedItemInfo.seriesIndex === index && focusedItemInfo.isFocusVisible
                                    ? focusedItemInfo.groupIndex
                                    : undefined, hoveredItemIndex: hoveredItemInfo?.isCurrent && hoveredItemInfo.seriesIndex === index
                                    ? hoveredItemInfo.groupIndex
                                    : undefined, isSelectionEnabled: selectionMode === 'single' || selectionMode === 'multiple', isDrillEnabled: drilling === 'on' }));
                        })] }));
        } }));
}

export { LineAreaChart as L };
//# sourceMappingURL=LineAreaChart-8790fc26.js.map
