/* @oracle/oraclejet-preact: 16.0.1 */
import { jsxs, jsx } from 'preact/jsx-runtime';
import { useState, useEffect, useRef } from 'preact/hooks';
import { n as normalizePosition } from './positionUtils-c728c841.js';
import { u as useAnimation } from './useAnimation-7d188187.js';
import { u as useUser } from './useUser-6e653d84.js';
import { u as useTestId } from './useTestId-5050f73e.js';
import { u as useBreakpointValues } from './useBreakpointValues-37dca33f.js';
import { u as useComponentTheme } from './useComponentTheme-7202d4c1.js';
import { c as classNames } from './classNames-2e96f1d4.js';
import { DrawerLayoutRedwoodTheme } from './UNSAFE_DrawerLayout/themes/redwood/DrawerLayoutTheme.js';
import './TopLayerHost-d50e5624.js';
import 'preact';
import './Common/themes/redwood/theme.js';
import './Common/themes/themeContract.css.js';
import 'preact/compat';
import './useBreakpoints-b564ead5.js';
import './logger-c710d0a3.js';
import './UNSAFE_DrawerLayout/themes/DrawerLayoutStyles.css.js';
import './DrawerLayoutStyles.styles.css';
import './UNSAFE_DrawerLayout/themes/redwood/DrawerLayoutBaseTheme.css.js';
import './UNSAFE_DrawerLayout/themes/redwood/DrawerLayoutBaseTheme.styles.css';
import './UNSAFE_DrawerLayout/themes/redwood/DrawerLayoutVariants.css.js';
import './vanilla-extract-recipes-createRuntimeFn.esm-8c3e2104.js';

/**
 * A drawer layout adds expandable side contents (drawers) alongside some primary content.
 *
 * These drawers automatically swap 'reflow' and 'overlay' display mode based on width of the page and can be placed at the 'start', 'end' or 'bottom' edge.
 */
const DrawerLayout = ({ children, startDrawer, endDrawer, bottomDrawer, isStartOpen = false, isEndOpen = false, isBottomOpen = false, startDisplay, endDisplay, bottomDisplay, onClose, onTransitionEnd, testId // ...props
 }) => {
    // Constants
    const zero = '0px';
    const duration = 250;
    const stringStart = 'start';
    const stringEnd = 'end';
    const stringBottom = 'bottom';
    const stringInitial = 'initial';
    const stringOpening = 'opening';
    const stringOpen = 'open';
    const stringClosing = 'closing';
    const stringUnmounted = 'unmounted';
    const stringFullOverlay = 'full-overlay';
    const stringOverlay = 'overlay';
    const stringReflow = 'reflow';
    const stringHidden = 'hidden';
    const stringVisible = 'visible';
    // Initialize states
    const [startDrawerStatus, setStartDrawerStatus] = useState(isStartOpen ? stringInitial : stringUnmounted);
    const [endDrawerStatus, setEndDrawerStatus] = useState(isEndOpen ? stringInitial : stringUnmounted);
    const [bottomDrawerStatus, setBottomDrawerStatus] = useState(isBottomOpen ? stringInitial : stringUnmounted);
    const [startVisibility, setStartVisibility] = useState(stringHidden);
    const [endVisibility, setEndVisibility] = useState(stringHidden);
    const [bottomVisibility, setBottomVisibility] = useState(stringHidden);
    // Breakpoints:
    // sm: 0-600px (full-overlay)
    // md: 600-1024px (overlay)
    // lg: >1024px (reflow)
    const displayMode = useBreakpointValues({
        sm: 'full-overlay',
        md: 'overlay',
        lg: 'reflow'
    });
    const [viewportDisplayMode, setViewportDisplayMode] = useState(displayMode);
    useEffect(() => {
        // Size breakpoint crossed
        setViewportDisplayMode(displayMode);
    }, [displayMode]);
    // Refs
    const startDrawerRef = useRef(null);
    const endDrawerRef = useRef(null);
    const bottomDrawerRef = useRef(null);
    const getElementHeight = (element) => {
        return Math.round(element.offsetHeight);
    };
    const getElementWidth = (element) => {
        return Math.round(element.getBoundingClientRect().width);
    };
    // Normalizing position
    const { direction } = useUser();
    const isLtr = direction === 'ltr';
    const getNormalizedPlacement = (placement) => {
        return normalizePosition(placement, direction);
    };
    const getTranslateX = (placement, ref) => {
        return `${getNormalizedPlacement(placement) === 'right' ? '' : '-'}${getElementWidth(ref.current)}px`;
    };
    const getTranslateY = (ref) => {
        return `${getElementHeight(ref.current)}px`;
    };
    const isBottomPlacement = (placement) => {
        return placement === stringBottom;
    };
    const isAnimationPending = (status) => {
        return [stringOpening, stringClosing].indexOf(status) > -1;
    };
    const isDrawerAnimationPending = (placement) => {
        switch (placement) {
            case stringStart:
                return isAnimationPending(startDrawerStatus);
            case stringEnd:
                return isAnimationPending(endDrawerStatus);
            case stringBottom:
                return isAnimationPending(bottomDrawerStatus);
        }
    };
    const isOverlayDisplayMode = (displayMode) => {
        return [stringFullOverlay, stringOverlay].indexOf(displayMode) > -1;
    };
    const getIsDrawerOpen = (placement) => {
        switch (placement) {
            case stringStart:
                return isStartOpen;
            case stringEnd:
                return isEndOpen;
            case stringBottom:
                return isBottomOpen;
        }
    };
    const getDrawerVisibility = (placement) => {
        switch (placement) {
            case stringStart:
                return startVisibility;
            case stringEnd:
                return endVisibility;
            case stringBottom:
                return bottomVisibility;
        }
    };
    const setDrawerVisibility = (placement, visibility) => {
        switch (placement) {
            case stringStart:
                return setStartVisibility(visibility);
            case stringEnd:
                return setEndVisibility(visibility);
            case stringBottom:
                return setBottomVisibility(visibility);
        }
    };
    const getDrawerRef = (placement) => {
        switch (placement) {
            case stringStart:
                return startDrawerRef;
            case stringEnd:
                return endDrawerRef;
            case stringBottom:
                return bottomDrawerRef;
        }
    };
    const getAnimatedNodeRef = (placement) => {
        switch (placement) {
            case stringStart:
                return startAnimatedNodeRef;
            case stringEnd:
                return endAnimatedNodeRef;
            case stringBottom:
                return bottomAnimatedNodeRef;
        }
    };
    const getDrawerStatus = (placement) => {
        switch (placement) {
            case stringStart:
                return startDrawerStatus;
            case stringEnd:
                return endDrawerStatus;
            case stringBottom:
                return bottomDrawerStatus;
        }
    };
    const setDrawerStatus = (placement, status) => {
        switch (placement) {
            case stringStart:
                setStartDrawerStatus(status);
                break;
            case stringEnd:
                setEndDrawerStatus(status);
                break;
            case stringBottom:
                setBottomDrawerStatus(status);
                break;
        }
    };
    const resetAnimatedNodeRef = (placement) => {
        switch (placement) {
            case stringStart:
                startAnimatedNodeRef(null);
                break;
            case stringEnd:
                endAnimatedNodeRef(null);
                break;
            case stringBottom:
                bottomAnimatedNodeRef(null);
                break;
        }
    };
    const getDrawerDisplayProp = (placement) => {
        switch (placement) {
            case stringStart:
                return startDisplay;
            case stringEnd:
                return endDisplay;
            case stringBottom:
                return bottomDisplay;
        }
    };
    const getDrawerDisplayMode = (placement) => {
        const displayMode = getDrawerDisplayProp(placement);
        return displayMode || viewportDisplayMode;
    };
    // Animation config
    const getAnimationStates = (placement, ref) => {
        const bottomPlacement = isBottomPlacement(placement);
        if (getDrawerDisplayMode(placement) === stringReflow) {
            // Reflow display mode
            // Effect: Expand/Collapse
            if (bottomPlacement) {
                return {
                    opening: () => {
                        return {
                            from: {
                                maxHeight: zero,
                                overflowX: 'hidden'
                            },
                            to: {
                                maxHeight: `${getElementHeight(ref.current)}px`,
                                overflowX: 'hidden'
                            },
                            options: {
                                duration: duration
                            }
                        };
                    },
                    closing: () => {
                        return {
                            from: {
                                maxHeight: `${getElementHeight(ref.current)}px`
                            },
                            to: {
                                maxHeight: zero,
                                overflowX: 'hidden'
                            },
                            options: {
                                duration: duration
                            }
                        };
                    }
                };
            }
            else {
                return {
                    opening: () => {
                        return {
                            from: {
                                maxWidth: zero,
                                overflowY: 'hidden'
                            },
                            to: {
                                maxWidth: `${getElementWidth(ref.current)}px`,
                                overflowY: 'hidden'
                            },
                            options: {
                                duration: duration
                            }
                        };
                    },
                    closing: () => {
                        return {
                            from: {
                                maxWidth: `${getElementWidth(ref.current)}px`
                            },
                            to: {
                                maxWidth: zero,
                                overflowY: 'hidden'
                            },
                            options: {
                                duration: duration
                            }
                        };
                    }
                };
            }
        }
        else {
            // FullOverlay or Overlay display mode
            // Effect: Slidein/Slideout
            if (bottomPlacement) {
                return {
                    opening: () => {
                        return {
                            from: {
                                translateY: getTranslateY(ref)
                            },
                            to: {
                                translateY: zero
                            },
                            options: {
                                duration: duration
                            }
                        };
                    },
                    closing: () => {
                        return {
                            from: {
                                translateY: zero
                            },
                            to: {
                                translateY: getTranslateY(ref)
                            },
                            options: {
                                duration: duration
                            }
                        };
                    }
                };
            }
            else {
                return {
                    opening: () => {
                        return {
                            from: {
                                translateX: getTranslateX(placement, ref)
                            },
                            to: {
                                translateX: zero
                            },
                            options: {
                                duration: duration
                            }
                        };
                    },
                    closing: () => {
                        return {
                            from: {
                                translateX: zero
                            },
                            to: {
                                translateX: getTranslateX(placement, ref)
                            },
                            options: {
                                duration: duration
                            }
                        };
                    }
                };
            }
        }
    };
    // Custom hook to be shared for each side
    // It can't be a function as a hook can only be called within a custom hook
    function useNodeRef(placement, ref) {
        const isDrawerOpen = getIsDrawerOpen(placement);
        const status = getDrawerStatus(placement);
        return useAnimation(status, {
            animationStates: getAnimationStates(placement, ref),
            isAnimatedOnMount: true,
            onAnimationEnd: () => {
                if (status === stringOpening) {
                    setDrawerStatus(placement, stringOpen);
                    onTransitionEnd?.({
                        placement: placement,
                        value: true
                    });
                }
                else if (!isDrawerOpen) {
                    resetAnimatedNodeRef(placement);
                    setDrawerVisibility(placement, stringHidden);
                    setDrawerStatus(placement, stringUnmounted);
                    onTransitionEnd?.({
                        placement: placement,
                        value: false
                    });
                }
            }
        }).nodeRef;
    }
    // Drawer refs
    const startAnimatedNodeRef = useNodeRef(stringStart, startDrawerRef);
    const endAnimatedNodeRef = useNodeRef(stringEnd, endDrawerRef);
    const bottomAnimatedNodeRef = useNodeRef(stringBottom, bottomDrawerRef);
    //testId support
    const testIdProps = useTestId(testId);
    // Custom hook to be shared for each side
    // It can't be a function as a hook can only be called within a custom hook
    function useDrawerStatus(placement) {
        const isDrawerOpen = getIsDrawerOpen(placement);
        const drawerStatus = getDrawerStatus(placement);
        useEffect(() => {
            // 1. Ignore status update on initial render
            if (drawerStatus === stringUnmounted && !isDrawerOpen) {
                return;
            }
            // 2. Mount visually hidden component
            // Component gets mounted, but will be visually hidden to avoid flickering
            // Flickering happens because it gets rendered before animation hook repositions
            // it in next render cycle
            if (drawerStatus === stringUnmounted && isDrawerOpen) {
                setDrawerStatus(placement, stringInitial);
                return;
            }
            // 3. Setup animation
            // Component has been mounted.
            // We will kick out animation as node is already present in DOM.
            // To avoid flickering, we still keep it visually hidden until the next render cycle.
            if (drawerStatus === stringInitial && isDrawerOpen) {
                const drawerRef = getDrawerRef(placement);
                const animatedNodeRef = getAnimatedNodeRef(placement);
                animatedNodeRef(drawerRef.current);
                setDrawerStatus(placement, stringOpening);
                return;
            }
            // 4. Unhide component. It becomes visible.
            if (drawerStatus === stringOpening && isDrawerOpen) {
                setDrawerVisibility(placement, stringVisible);
                return;
            }
            // 5. Status 'open' is set in 'onAnimationEnd' handler
            // 6. Closing
            if (!isDrawerOpen) {
                setDrawerStatus(placement, stringClosing);
            }
        }, [isDrawerOpen, drawerStatus, placement]);
    }
    // Drawers' status handler±
    useDrawerStatus(stringStart);
    useDrawerStatus(stringEnd);
    useDrawerStatus(stringBottom);
    // Keydown handlers
    const handleKeyDownStart = (event) => {
        handleKeyDown(stringStart, event);
    };
    const handleKeyDownEnd = (event) => {
        handleKeyDown(stringEnd, event);
    };
    const handleKeyDownBottom = (event) => {
        handleKeyDown(stringBottom, event);
    };
    const handleKeyDown = (placement, event) => {
        if (event.code === 'Escape') {
            // Handle only in overlay mode
            if (isOverlayDisplayMode(getDrawerDisplayMode(placement))) {
                if (event.code === 'Escape') {
                    onClose?.({
                        placement,
                        reason: 'escapeKey'
                    });
                }
            }
        }
    };
    // Styles
    const { classes, styles } = useComponentTheme(DrawerLayoutRedwoodTheme);
    const getVisibilityClass = (visibility) => {
        return visibility === stringHidden ? styles.visibilityHidden : styles.visibilityVisible;
    };
    const getClassNames = (placement) => {
        const displayMode = getDrawerDisplayMode(placement);
        const visibility = getDrawerVisibility(placement);
        const visibilityClass = getVisibilityClass(visibility);
        const animationPendingClass = isDrawerAnimationPending(placement) && styles.animatedDrawer;
        switch (placement) {
            case stringStart:
                if (displayMode === stringFullOverlay) {
                    return classNames([
                        styles.overlayDrawer,
                        styles.fullOverlayDrawerSide,
                        isLtr && styles.overlayDrawerLeft,
                        !isLtr && styles.overlayDrawerRight,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
                else if (displayMode === stringOverlay) {
                    return classNames([
                        styles.overlayDrawer,
                        styles.basicOverlayDrawerSide,
                        isLtr && styles.overlayDrawerLeft,
                        !isLtr && styles.overlayDrawerRight,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
                else {
                    return classNames([
                        styles.reflowDrawer,
                        isLtr && styles.reflowDrawerLeft,
                        !isLtr && styles.reflowDrawerRight,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
            case stringEnd:
                if (displayMode === stringFullOverlay) {
                    return classNames([
                        styles.overlayDrawer,
                        styles.fullOverlayDrawerSide,
                        isLtr && styles.overlayDrawerRight,
                        !isLtr && styles.overlayDrawerLeft,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
                else if (displayMode === stringOverlay) {
                    return classNames([
                        styles.overlayDrawer,
                        styles.basicOverlayDrawerSide,
                        isLtr && styles.overlayDrawerRight,
                        !isLtr && styles.overlayDrawerLeft,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
                else {
                    return classNames([
                        styles.reflowDrawer,
                        isLtr && styles.reflowDrawerRight,
                        !isLtr && styles.reflowDrawerLeft,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
            case stringBottom:
                if ([stringOverlay, stringFullOverlay].indexOf(displayMode) > -1) {
                    return classNames([styles.overlayDrawer, styles.overlayDrawerBottom, visibilityClass]);
                }
                else {
                    return classNames([
                        styles.reflowDrawer,
                        styles.reflowDrawerBottom,
                        visibilityClass,
                        animationPendingClass
                    ]);
                }
        }
    };
    const getDrawer = (placement) => {
        const isMounted = getDrawerStatus(placement) !== stringUnmounted;
        if (isMounted) {
            switch (placement) {
                case stringStart:
                    return (jsx("div", { ref: startDrawerRef, className: getClassNames(stringStart), tabIndex: isOverlayDisplayMode(getDrawerDisplayMode(stringStart)) ? -1 : undefined, onKeyDown: handleKeyDownStart, children: startDrawer }));
                case stringEnd:
                    return (jsx("div", { ref: endDrawerRef, className: getClassNames(stringEnd), tabIndex: isOverlayDisplayMode(getDrawerDisplayMode(stringStart)) ? -1 : undefined, onKeyDown: handleKeyDownEnd, children: endDrawer }));
                case stringBottom:
                    return (jsx("div", { ref: bottomDrawerRef, className: getClassNames(stringBottom), tabIndex: isOverlayDisplayMode(getDrawerDisplayMode(stringStart)) ? -1 : undefined, onKeyDown: handleKeyDownBottom, children: bottomDrawer }));
            }
        }
        return null;
    };
    return (jsxs("div", { className: classes, ...testIdProps, children: [getDrawer(stringStart), jsxs("div", { className: styles.centerSection, children: [jsx("div", { className: styles.mainContent, children: children }), getDrawer(stringBottom)] }), getDrawer(stringEnd)] }));
};

export { DrawerLayout };
//# sourceMappingURL=UNSAFE_DrawerLayout.js.map
