/* @oracle/oraclejet-preact: 16.0.1 */
/**
 * @license
 * Copyright (c) %FIRST_YEAR% %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
/**
 * Returns the aria label for the chart item
 * @param translations The translations bundle.
 * @param series The series of the item.
 * @param group The group of the item.
 * @param value The value of the item.
 * @param isSelected Whether the item is selected.
 * @param isDrillable Whether the item is drillable.
 * @param accessibleLabel The accessible label of the item.
 * @returns
 */
function getItemAriaLabel(translations, series, group, value, isSelectionEnabled, isSelected, isDrillable, accessibleLabel) {
    const seriesLabel = translations.chart_labelSeries();
    const groupLabel = translations.chart_labelGroup();
    const valueLabel = translations.chart_labelValue();
    const label = accessibleLabel || `${seriesLabel}: ${series}; ${groupLabel}: ${group}; ${valueLabel}: ${value};`;
    const selectedLabel = translations.dataVisualization_stateSelected();
    const unselectedLabel = translations.dataVisualization_stateUnselected();
    const drillableLabel = translations.viz_drillable();
    return [
        label,
        isSelectionEnabled ? (isSelected ? selectedLabel : unselectedLabel) : '',
        isDrillable ? drillableLabel : ''
    ].filter(Boolean).join(' ');
}

/**
 * @license
 * Copyright (c) %FIRST_YEAR% %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
/**
 * Returns the next navigable item in the same series.
 * @param seriesIndex
 * @param groupIndex
 * @param numOfGroups
 * @returns
 */
function getNextItemInSeries(info, viewportEndIndex, isNavigable) {
    const { seriesIndex, groupIndex } = info;
    const nextGroupIndex = groupIndex + 1;
    if (nextGroupIndex > viewportEndIndex) {
        return;
    }
    if (isNavigable(seriesIndex, nextGroupIndex)) {
        return { seriesIndex, groupIndex: nextGroupIndex };
    }
    return getNextItemInSeries({
        seriesIndex,
        groupIndex: nextGroupIndex
    }, viewportEndIndex, isNavigable);
}
/**
 * Returns the previous navigable item in the same series.
 * @param seriesIndex
 * @param itemIndex
 * @returns
 */
function getPrevItemInSeries(info, viewportStartIndex, isNavigable) {
    const { seriesIndex, groupIndex } = info;
    const prevGroupIndex = groupIndex - 1;
    if (prevGroupIndex < viewportStartIndex) {
        return;
    }
    if (isNavigable(seriesIndex, prevGroupIndex)) {
        return { seriesIndex, groupIndex: prevGroupIndex };
    }
    if (prevGroupIndex === 0) {
        return undefined;
    }
    return getPrevItemInSeries({ seriesIndex, groupIndex: prevGroupIndex }, viewportStartIndex, isNavigable);
}
/**
 * Returns the previous navigable item in the same group.
 * @param seriesIndex
 * @param groupIndex
 * @param isNavigable
 */
function getPrevItemInGroup(info, isNavigable) {
    const { seriesIndex, groupIndex } = info;
    const prevSeriesIndex = seriesIndex - 1;
    if (prevSeriesIndex < 0) {
        return;
    }
    if (isNavigable(prevSeriesIndex, groupIndex)) {
        return { seriesIndex: prevSeriesIndex, groupIndex };
    }
    return getPrevItemInGroup({
        seriesIndex: prevSeriesIndex,
        groupIndex
    }, isNavigable);
}
/**
 * Returns the next navigable item in the same group.
 * @param seriesIndex
 * @param groupIndex
 * @param numOfSeries
 * @param numOfGroups
 */
function getNextItemInGroup(info, numOfSeries, isNavigable) {
    const { seriesIndex, groupIndex } = info;
    const nextSeriesIndex = seriesIndex + 1;
    if (nextSeriesIndex >= numOfSeries) {
        return;
    }
    if (isNavigable(nextSeriesIndex, groupIndex)) {
        return { seriesIndex: nextSeriesIndex, groupIndex };
    }
    return getNextItemInGroup({
        seriesIndex: nextSeriesIndex,
        groupIndex
    }, numOfSeries, isNavigable);
}
/**
 * Returns the next item in the group with higher value than the current item.
 * @param info The ItemInfo of current item.
 * @param numSeries The number of series in the group.
 * @param getDataItem The function that returns the chart data item.
 * @returns
 */
function getNextHigherItemInGroup(info, numSeries, getDataItem, isItemNavigable) {
    const { groupIndex, seriesIndex } = info;
    const item = getDataItem(seriesIndex, groupIndex);
    if (!item) {
        return info;
    }
    let minValue = Number.MAX_VALUE;
    let minIndex = seriesIndex;
    for (let i = 0; i < numSeries; i++) {
        const itemValue = getDataItem(i, groupIndex)?.value;
        if (!itemValue ||
            itemValue < item.value ||
            !isItemNavigable(i, groupIndex) ||
            seriesIndex === i) {
            continue;
        }
        if (itemValue < minValue) {
            minValue = itemValue;
            minIndex = i;
        }
    }
    return { seriesIndex: minIndex, groupIndex };
}
/**
 * Returns the item info of the item in the group with lower value than the current item.
 * @param info The ItemInfo of the current item.
 * @param numSeries The number of series in the group.
 * @param getDataItem The function that returns the chart data item.
 * @returns
 */
function getNextLowerItemInGroup(info, numSeries, getDataItem, isItemNavigable) {
    const { groupIndex, seriesIndex } = info;
    const item = getDataItem(seriesIndex, groupIndex);
    if (!item) {
        return info;
    }
    let maxValue = Number.MIN_VALUE;
    let maxIndex = seriesIndex;
    for (let i = 0; i < numSeries; i++) {
        const itemValue = getDataItem(i, groupIndex)?.value;
        if (!itemValue ||
            itemValue > item.value ||
            !isItemNavigable(i, groupIndex) ||
            seriesIndex === i) {
            continue;
        }
        if (itemValue > maxValue) {
            maxValue = itemValue;
            maxIndex = i;
        }
    }
    return { seriesIndex: maxIndex, groupIndex };
}
function getLineAreaChartNavUtil(getDataItem, numSeries, viewportStartIndex, viewportEndIndex, isStacked, isRtl, isHoriz) {
    const isItemNavigable = (seriesIndex, groupIndex) => {
        const barItem = getDataItem(seriesIndex, groupIndex);
        return Boolean(barItem);
    };
    const getNextChartItem = (itemInfo, arrow) => {
        let newInfo;
        const info = {
            groupIndex: itemInfo.groupIndex,
            seriesIndex: itemInfo.seriesIndex
        };
        if (info.groupIndex < viewportStartIndex) {
            info.groupIndex = viewportStartIndex;
            return info;
        }
        else if (info.groupIndex > viewportEndIndex) {
            info.groupIndex = viewportEndIndex;
            return info;
        }
        if (arrow === 'ArrowLeft') {
            if (!isHoriz) {
                newInfo = isRtl
                    ? getNextItemInSeries(info, viewportEndIndex, isItemNavigable)
                    : getPrevItemInSeries(info, viewportStartIndex, isItemNavigable);
            }
            else {
                newInfo = isRtl
                    ? getNextHigherItemInGroup(info, numSeries, getDataItem, isItemNavigable)
                    : getNextLowerItemInGroup(info, numSeries, getDataItem, isItemNavigable);
            }
        }
        else if (arrow === 'ArrowRight') {
            if (!isHoriz) {
                newInfo = isRtl
                    ? getPrevItemInSeries(info, viewportStartIndex, isItemNavigable)
                    : getNextItemInSeries(info, viewportEndIndex, isItemNavigable);
            }
            else {
                newInfo = !isRtl
                    ? getNextHigherItemInGroup(info, numSeries, getDataItem, isItemNavigable)
                    : getNextLowerItemInGroup(info, numSeries, getDataItem, isItemNavigable);
            }
        }
        else if (arrow === 'ArrowDown') {
            if (isHoriz) {
                newInfo = getPrevItemInSeries(info, viewportStartIndex, isItemNavigable);
            }
            else {
                if (isStacked) {
                    newInfo = getPrevItemInGroup(info, isItemNavigable);
                }
                else {
                    newInfo = getNextLowerItemInGroup(info, numSeries, getDataItem, isItemNavigable);
                }
            }
        }
        else {
            if (isHoriz) {
                newInfo = getNextItemInSeries(info, viewportEndIndex, isItemNavigable);
            }
            else {
                if (isStacked) {
                    newInfo = getNextItemInGroup(info, numSeries, isItemNavigable);
                }
                else {
                    newInfo = getNextHigherItemInGroup(info, numSeries, getDataItem, isItemNavigable);
                }
            }
        }
        return newInfo || info;
    };
    return getNextChartItem;
}

export { getNextItemInSeries as a, getPrevItemInSeries as b, getPrevItemInGroup as c, getNextItemInGroup as d, getLineAreaChartNavUtil as e, getItemAriaLabel as g };
//# sourceMappingURL=navUtils-4a56ee79.js.map
