/**
 * @license
 * Copyright (c) %FIRST_YEAR% %CURRENT_YEAR%, Oracle and/or its affiliates.
 * Licensed under The Universal Permissive License (UPL), Version 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 * @ignore
 */
import { DataAxisProps, Group, GroupAxisProps, TimeAxisProps } from '../UNSAFE_Axis/axis.types';
import { ComponentChildren, VNode } from 'preact';
import { ChartPlotArea } from '../UNSAFE_PlotArea/PlotArea.types';
import { DataCursorProps } from '../hooks/UNSAFE_useChartDataCursor';
import type { TestIdProps } from '../hooks/UNSAFE_useTestId';
import { Property } from 'csstype';
export type Dimension = {
    x: number;
    y: number;
    height: number;
    width: number;
};
export type GroupInfo = {
    groupIndex: number;
    isCurrent?: boolean;
    isFocusVisible?: boolean;
};
export type Info = {
    seriesIndex?: number;
} & Partial<GroupInfo>;
export type ItemInfo = Required<Pick<Info, 'seriesIndex' | 'groupIndex'>>;
export type AxisPosition = 'start' | 'end' | 'top' | 'bottom';
export type viewportDetail = {
    /**
     * The viewport min of the x axis.
     */
    xMin?: number;
    /**
     * The viewport max of the x axis.
     */
    xMax?: number;
    /**
     * The viewport min of the y axis.
     */
    yMin?: number;
    /**
     * The viewport max of the y axis.
     */
    yMax?: number;
};
export type ItemContext<K, D> = {
    /**
     * The group index of the given item.
     */
    groupIndex: number;
    /**
     * The series index of the given item.
     */
    seriesIndex: number;
    /**
     * The id of the given item.
     */
    id: K;
    /**
     * The data item.
     */
    data: D;
};
export type GroupContext = {
    /**
     * The index of the group in the groups array.
     */
    groupIndex: number;
    /**
     * The id of the group.
     */
    id: string | number;
    /**
     * The group data.
     */
    group: Group;
};
export type SelectionProps = {
    /**
     * An array used to define the ids of the initially selected objects.
     */
    selectedIds?: (string | number)[];
    /**
     * The type of selection behavior that is enabled on the chart.
     * If single or multiple is specified, selection gestures will be enabled, and the chart's selection styling will be applied to all items specified by the selection attribute.
     * If none is specified, selection gestures will be disabled, and the chart's selection styling will not be applied to any items specified by the selection attribute.
     */
    selectionMode?: 'single' | 'multiple' | 'none';
    /**
     * The callback to be fired when chart selection is changed. selection prop should be updated by this function.
     */
    onSelectionChange?: (detail: {
        ids: (string | number)[];
    }) => void;
};
export type DrillProps<K, D> = {
    /**
     * Triggered on a chart item (double click if selection is enabled, single click otherwise).
     */
    onItemDrill?: (detail: ItemContext<K, D>) => void;
    /**
     * Triggered on a chart group drill gesture (double click if selection is enabled, single click otherwise).
     */
    onGroupDrill?: (detail: GroupContext) => void;
    /**
     * Whether drilling is enabled.
     * Drillable objects will show a pointer cursor on hover and fire an onItemDrill or onGroupDrill event on click (double click if selection is enabled).
     * Use "on" to enable drilling for all group objects (x-axis labels) and data items such as bars, area and marker.
     * To enable or disable drilling on group or data item use the drilling attribute in each series, group, or data item.
     */
    drilling?: 'on' | 'off' | 'groupsOnly';
};
export type TextValueFormat = {
    /**
     * Whether the string is displayed in the tooltip.
     */
    isDisplayed?: boolean;
    /**
     * A string representing the label that is displayed before the value in the tooltip.
     */
    label?: string;
};
export type NumericalValueFormat = {
    /**
     * Whether the value is displayed in the tooltip.
     */
    isDisplayed?: boolean;
    /**
     * A string representing the label that is displayed before the value in the tooltip.
     */
    label?: string;
    /**
     * The function called to get the formatted value.
     */
    format?: (value: number) => string;
};
export type ValueFormats = {
    /**
     * Object that specifies the value formatting and tooltip behavior for the group.
     */
    group: TextValueFormat;
    /**
     * Object that specifies the value formatting and tooltip behavior for the series.
     */
    series: TextValueFormat;
    /**
     * Object that specifies the value formatting and tooltip behavior for the item value.
     */
    value: NumericalValueFormat;
};
export type ChartItem<K> = {
    /**
     * The  color of the bar item.
     */
    color?: Property.BorderColor;
    /**
     * The border color of the bar item.
     */
    borderColor?: Property.BorderColor;
    /**
     * The label to be used for accessibility purposes.
     */
    accessibleLabel?: string;
    /**
     * The value of the data item. If undefined data item will not be rendered.
     */
    value: number;
    /**
     * Defines the label for the data item.
     */
    label?: string;
    /**
     * The id of the data item.
     */
    id: K;
    /**
     * The x value of the data item.
     */
    x?: string;
    /**
     * Whether the item is drillable or not.
     */
    drilling?: 'on' | 'off' | 'inherit';
};
export type ChartSeries<K, D> = {
    /**
     * The id of the series.
     */
    id: K;
    /**
     * The name of the series.
     */
    name?: string;
    /**
     * The color of the items of the series.
     */
    color?: Property.BorderColor;
    /**
     * The border color of the items of the series.
     */
    borderColor?: Property.BorderColor;
    /**
     * The label to be used for accessibility purposes.
     */
    accessibleLabel?: string;
    /**
     * The items present in the sereis.
     */
    items: D[];
};
export type ChartProps<K, D> = {
    /**
     * Defines the hide and show behavior.When data items are hidden, the y axes can be optionally rescaled to fit to the remaining data
     */
    hideAndShowBehavior?: 'withoutRescale' | 'withRescale' | 'none';
    /**
     * The ids of the hidden items.
     */
    hiddenIds?: K[];
    /**
     * An array of ids used for highlighting. Series or data items matching the ids in the array will be highlighted.
     */
    highlightedIds?: K[];
    /**
     * The orientation of the chart.
     */
    orientation?: 'vertical' | 'horizontal';
    /**
     * The width of the chart.
     */
    width?: string;
    /**
     * Defines the behavior applied when hovering over data items. When hover behavior is set to dim it will dim all other data items except the one being hovered.
     */
    hoverBehavior?: 'dim' | 'none';
    /**
     * The viewport change callback.
     */
    onViewportChange?: (detail: viewportDetail) => void;
    /**
     * Triggered when item is hovered or keyboard focused.
     */
    onItemInput?: (detail: {
        seriesIndex?: number;
        groupIndex?: number;
    }) => void;
    /**
     * The function called to get the data item for given seriesIndex and groupIndex.
     */
    getDataItem: (seriesIndex: number, groupIndex: number) => D;
    /**
     * The height of the chart.
     */
    height?: string;
    /**
     * An object defining properties for the axis, tick marks, tick labels, and axis titles.
     */
    yAxis?: DataAxisProps;
    /**
     * An object defining properties for the xAxis, tick marks, tick labels, and axis titles.
     */
    xAxis?: GroupAxisProps | TimeAxisProps;
    /**
     * An array of groups identifying the group labels.
     */
    groups: Group[];
    /**
     * Whether the chart is stacked or not.
     */
    isStacked?: boolean;
    /**
     * The function that returns datatip for the given item context.
     */
    datatip?: (detail: ItemContext<K, D>) => {
        content: ComponentChildren;
        borderColor?: Property.BorderColor;
    };
    /**
     * The plot area of the chart.
     */
    plotArea?: ChartPlotArea;
    /**
     * Specifies the zoom and scroll behavior of the chart. "live" behavior means that the chart will be updated continuously as it is being manipulated.
     */
    zoomAndScroll?: 'live' | 'off';
    /**
     * An object specifying value formatting and tooltip behavior, whose keys generally correspond to the attribute names on the data items.
     */
    valueFormats?: ValueFormats;
    /**
     * The aria label of the chart.
     */
    'aria-label'?: string;
    /**
     * One or more ids (separated by spaces) of elements that label the chart. This will be used for the aria-labelledby of the DOM element.
     */
    'aria-labelledby'?: string;
    /**
     * One or more ids (separated by spaces) of elements that describe the chart. This will be used for the aria-describedby of the DOM element.
     */
    'aria-describedby'?: string;
} & TestIdProps & DataCursorProps & SelectionProps & DrillProps<K, D>;
export type getNextChartItemType = (info: Info, key: 'ArrowUp' | 'ArrowDown' | 'ArrowRight' | 'ArrowLeft') => Info;
export type getDataPosType = (seriesIndex: number, groupIndex: number) => Dimension | undefined;
export type Scale = {
    transform: (value: number) => number;
    invert: (value: number) => number;
    range: () => number[];
    domain: () => number[];
};
export type ChartCommonProps<K, D> = {
    getGapRatio: (startIndex: number, endIndex: number) => number;
    getNavUtil: (startIndex: number, endIndex: number) => getNextChartItemType;
    getDataItemPos: (xScale: Scale, yScale: Scale, groupWidth: number, gapRatio: number) => getDataPosType;
    findNearest: (xScale: Scale, yScale: Scale, groupWidth: number, gapRatio: number) => (x: number, y: number) => {
        seriesIndex: number;
        groupIndex: number;
        yCoord: number | undefined;
        xCoord: number;
    } | undefined;
    children: (context: {
        xStartIndex: number;
        xEndIndex: number;
        xScale: Scale;
        yScale: Scale;
        gapRatio: number;
        activeId?: string;
        focusedItemInfo: Info;
        hoveredItemInfo?: Info;
        groupWidth: number;
    }) => VNode;
    offset: number;
    series: ChartSeries<K, D>[];
    isRtl: boolean;
    renderGridLinesInFront: boolean;
} & Omit<ChartProps<K, D>, 'highlightedIds' | 'hoverBehavior'>;
